"""Tests for Scan Diff Reporter."""

from __future__ import annotations

import json

import pytest

from mcpaudit.scan_diff_reporter import (
    DiffReporter,
    DiffStatus,
    DiffSummary,
    ScanDiffEntry,
    Severity,
    TimeSeriesPoint,
    TimeSeriesReport,
    _extract_findings,
    _extract_scan_id,
    _extract_timestamp,
    _finding_id,
    format_json,
    format_markdown,
    format_time_series_json,
    format_time_series_markdown,
)


# ---------- Constants ----------

class TestDiffStatus:
    def test_new(self):
        assert DiffStatus.NEW == "NEW"

    def test_fixed(self):
        assert DiffStatus.FIXED == "FIXED"

    def test_unchanged(self):
        assert DiffStatus.UNCHANGED == "UNCHANGED"

    def test_modified(self):
        assert DiffStatus.MODIFIED == "MODIFIED"

    def test_all_contains_four(self):
        assert len(DiffStatus.ALL) == 4
        assert "NEW" in DiffStatus.ALL
        assert "FIXED" in DiffStatus.ALL


class TestSeverity:
    def test_critical(self):
        assert Severity.CRITICAL == "CRITICAL"

    def test_all_contains_five(self):
        assert len(Severity.ALL) == 5


# ---------- ScanDiffEntry ----------

class TestScanDiffEntry:
    def test_basic_creation(self):
        entry = ScanDiffEntry(finding_id="abc123", status=DiffStatus.NEW)
        assert entry.finding_id == "abc123"
        assert entry.status == "NEW"
        assert entry.severity == Severity.INFO

    def test_to_dict(self):
        entry = ScanDiffEntry(
            finding_id="abc123",
            status=DiffStatus.NEW,
            severity=Severity.HIGH,
            file="server.py",
            line=42,
            title="Unsafe input",
            rule_id="MCP-001",
        )
        d = entry.to_dict()
        assert d["finding_id"] == "abc123"
        assert d["status"] == "NEW"
        assert d["severity"] == "HIGH"
        assert d["file"] == "server.py"
        assert d["line"] == 42
        assert d["rule_id"] == "MCP-001"
        assert "previous_severity" not in d

    def test_to_dict_with_previous_severity(self):
        entry = ScanDiffEntry(
            finding_id="abc123",
            status=DiffStatus.MODIFIED,
            severity=Severity.CRITICAL,
            previous_severity=Severity.HIGH,
        )
        d = entry.to_dict()
        assert d["previous_severity"] == "HIGH"

    def test_default_values(self):
        entry = ScanDiffEntry(finding_id="x", status="NEW")
        assert entry.file == ""
        assert entry.line == 0
        assert entry.title == ""
        assert entry.rule_id == ""
        assert entry.previous_severity == ""


# ---------- DiffSummary ----------

class TestDiffSummary:
    def test_empty_summary(self):
        s = DiffSummary()
        assert s.new_count == 0
        assert s.fixed_count == 0
        assert s.unchanged_count == 0
        assert s.modified_count == 0
        assert s.net_delta == 0
        assert s.improvement_score == 0.0

    def test_counts(self):
        s = DiffSummary(entries=[
            ScanDiffEntry("a", DiffStatus.NEW, Severity.HIGH),
            ScanDiffEntry("b", DiffStatus.NEW, Severity.MEDIUM),
            ScanDiffEntry("c", DiffStatus.FIXED, Severity.LOW),
            ScanDiffEntry("d", DiffStatus.UNCHANGED, Severity.LOW),
            ScanDiffEntry("e", DiffStatus.MODIFIED, Severity.HIGH),
        ])
        assert s.new_count == 2
        assert s.fixed_count == 1
        assert s.unchanged_count == 1
        assert s.modified_count == 1

    def test_net_delta_positive(self):
        s = DiffSummary(entries=[
            ScanDiffEntry("a", DiffStatus.NEW, Severity.HIGH),
            ScanDiffEntry("b", DiffStatus.NEW, Severity.MEDIUM),
            ScanDiffEntry("c", DiffStatus.FIXED, Severity.LOW),
        ])
        assert s.net_delta == 1

    def test_net_delta_negative(self):
        s = DiffSummary(entries=[
            ScanDiffEntry("a", DiffStatus.FIXED, Severity.HIGH),
            ScanDiffEntry("b", DiffStatus.FIXED, Severity.MEDIUM),
            ScanDiffEntry("c", DiffStatus.NEW, Severity.LOW),
        ])
        assert s.net_delta == -1

    def test_net_delta_zero(self):
        s = DiffSummary(entries=[
            ScanDiffEntry("a", DiffStatus.NEW, Severity.HIGH),
            ScanDiffEntry("b", DiffStatus.FIXED, Severity.LOW),
        ])
        assert s.net_delta == 0

    def test_improvement_score_all_fixed(self):
        s = DiffSummary(entries=[
            ScanDiffEntry("a", DiffStatus.FIXED, Severity.CRITICAL),
        ])
        assert s.improvement_score == 100.0

    def test_improvement_score_all_new(self):
        s = DiffSummary(entries=[
            ScanDiffEntry("a", DiffStatus.NEW, Severity.HIGH),
        ])
        assert s.improvement_score == -100.0

    def test_improvement_score_mixed(self):
        s = DiffSummary(entries=[
            ScanDiffEntry("a", DiffStatus.NEW, Severity.HIGH),
            ScanDiffEntry("b", DiffStatus.FIXED, Severity.HIGH),
        ])
        assert s.improvement_score == 0.0

    def test_improvement_score_weighted(self):
        # Fixed critical (10) vs new low (1) -> (10-1)/(10+1)*100 = 81.82
        s = DiffSummary(entries=[
            ScanDiffEntry("a", DiffStatus.FIXED, Severity.CRITICAL),
            ScanDiffEntry("b", DiffStatus.NEW, Severity.LOW),
        ])
        assert s.improvement_score > 80

    def test_severity_breakdown(self):
        s = DiffSummary(entries=[
            ScanDiffEntry("a", DiffStatus.NEW, Severity.CRITICAL),
            ScanDiffEntry("b", DiffStatus.FIXED, Severity.HIGH),
            ScanDiffEntry("c", DiffStatus.FIXED, Severity.HIGH),
            ScanDiffEntry("d", DiffStatus.UNCHANGED, Severity.LOW),
        ])
        bd = s.severity_breakdown
        assert bd["CRITICAL"]["new"] == 1
        assert bd["CRITICAL"]["fixed"] == 0
        assert bd["HIGH"]["new"] == 0
        assert bd["HIGH"]["fixed"] == 2
        assert bd["LOW"]["new"] == 0
        assert bd["LOW"]["fixed"] == 0

    def test_to_dict(self):
        s = DiffSummary(
            scan_before_id="scan1",
            scan_after_id="scan2",
            timestamp_before="2024-01-01",
            timestamp_after="2024-01-02",
            entries=[ScanDiffEntry("a", DiffStatus.NEW, Severity.HIGH)],
        )
        d = s.to_dict()
        assert d["scan_before_id"] == "scan1"
        assert d["scan_after_id"] == "scan2"
        assert d["counts"]["new"] == 1
        assert d["net_delta"] == 1
        assert len(d["entries"]) == 1


# ---------- TimeSeriesPoint ----------

class TestTimeSeriesPoint:
    def test_creation(self):
        p = TimeSeriesPoint(timestamp="2024-01-01", total_findings=5)
        assert p.timestamp == "2024-01-01"
        assert p.total_findings == 5

    def test_to_dict(self):
        p = TimeSeriesPoint(
            timestamp="t1", total_findings=10,
            new_count=3, fixed_count=2,
            net_delta=1, improvement_score=-20.0,
        )
        d = p.to_dict()
        assert d["timestamp"] == "t1"
        assert d["new_count"] == 3
        assert d["net_delta"] == 1


# ---------- TimeSeriesReport ----------

class TestTimeSeriesReport:
    def test_empty(self):
        r = TimeSeriesReport()
        assert r.points == []
        assert r.overall_trend == ""

    def test_to_dict(self):
        r = TimeSeriesReport(
            points=[TimeSeriesPoint("t1")],
            overall_trend="improving",
            total_new_across_all=5,
            total_fixed_across_all=10,
        )
        d = r.to_dict()
        assert len(d["points"]) == 1
        assert d["overall_trend"] == "improving"


# ---------- Helper Functions ----------

class TestFindingId:
    def test_deterministic(self):
        f = {"rule_id": "R1", "file": "a.py", "title": "bug"}
        assert _finding_id(f) == _finding_id(f)

    def test_different_findings_different_ids(self):
        f1 = {"rule_id": "R1", "file": "a.py", "title": "bug1"}
        f2 = {"rule_id": "R1", "file": "a.py", "title": "bug2"}
        assert _finding_id(f1) != _finding_id(f2)

    def test_empty_finding(self):
        fid = _finding_id({})
        assert isinstance(fid, str)
        assert len(fid) == 16

    def test_uses_description_fallback(self):
        f = {"rule_id": "R1", "file": "a.py", "description": "desc"}
        fid = _finding_id(f)
        assert len(fid) == 16


class TestExtractFindings:
    def test_results_format(self):
        data = {"results": [{"findings": [{"rule_id": "R1"}, {"rule_id": "R2"}]}]}
        assert len(_extract_findings(data)) == 2

    def test_findings_format(self):
        data = {"findings": [{"rule_id": "R1"}]}
        assert len(_extract_findings(data)) == 1

    def test_empty_data(self):
        assert _extract_findings({}) == []

    def test_multiple_results(self):
        data = {"results": [
            {"findings": [{"rule_id": "R1"}]},
            {"findings": [{"rule_id": "R2"}, {"rule_id": "R3"}]},
        ]}
        assert len(_extract_findings(data)) == 3


class TestExtractTimestamp:
    def test_top_level(self):
        assert _extract_timestamp({"timestamp": "2024-01-01"}) == "2024-01-01"

    def test_in_summary(self):
        data = {"summary": {"timestamp": "2024-02-01"}}
        assert _extract_timestamp(data) == "2024-02-01"

    def test_missing(self):
        assert _extract_timestamp({}) == ""


class TestExtractScanId:
    def test_explicit_id(self):
        assert _extract_scan_id({"scan_id": "abc"}) == "abc"

    def test_generated_id(self):
        sid = _extract_scan_id({"findings": []})
        assert isinstance(sid, str)
        assert len(sid) == 12


# ---------- DiffReporter ----------

class TestDiffReporter:
    def setup_method(self):
        self.reporter = DiffReporter()

    def test_empty_scans(self):
        s = self.reporter.compare({}, {})
        assert s.new_count == 0
        assert s.fixed_count == 0

    def test_all_new(self):
        before = {"findings": []}
        after = {"findings": [
            {"rule_id": "R1", "file": "a.py", "title": "Bug", "severity": "HIGH"},
        ]}
        s = self.reporter.compare(before, after)
        assert s.new_count == 1
        assert s.fixed_count == 0

    def test_all_fixed(self):
        before = {"findings": [
            {"rule_id": "R1", "file": "a.py", "title": "Bug", "severity": "HIGH"},
        ]}
        after = {"findings": []}
        s = self.reporter.compare(before, after)
        assert s.new_count == 0
        assert s.fixed_count == 1

    def test_unchanged(self):
        finding = {"rule_id": "R1", "file": "a.py", "title": "Bug", "severity": "HIGH"}
        s = self.reporter.compare({"findings": [finding]}, {"findings": [finding]})
        assert s.unchanged_count == 1
        assert s.new_count == 0
        assert s.fixed_count == 0

    def test_modified_severity(self):
        before = {"findings": [
            {"rule_id": "R1", "file": "a.py", "title": "Bug", "severity": "HIGH"},
        ]}
        after = {"findings": [
            {"rule_id": "R1", "file": "a.py", "title": "Bug", "severity": "CRITICAL"},
        ]}
        s = self.reporter.compare(before, after)
        assert s.modified_count == 1
        mod = [e for e in s.entries if e.status == DiffStatus.MODIFIED][0]
        assert mod.previous_severity == "HIGH"
        assert mod.severity == "CRITICAL"

    def test_mixed_changes(self):
        before = {"findings": [
            {"rule_id": "R1", "file": "a.py", "title": "Old bug", "severity": "LOW"},
            {"rule_id": "R2", "file": "b.py", "title": "Stays", "severity": "MEDIUM"},
        ]}
        after = {"findings": [
            {"rule_id": "R2", "file": "b.py", "title": "Stays", "severity": "MEDIUM"},
            {"rule_id": "R3", "file": "c.py", "title": "New bug", "severity": "HIGH"},
        ]}
        s = self.reporter.compare(before, after)
        assert s.new_count == 1
        assert s.fixed_count == 1
        assert s.unchanged_count == 1

    def test_timestamps_extracted(self):
        before = {"timestamp": "2024-01-01", "findings": []}
        after = {"timestamp": "2024-01-02", "findings": []}
        s = self.reporter.compare(before, after)
        assert s.timestamp_before == "2024-01-01"
        assert s.timestamp_after == "2024-01-02"

    def test_scan_ids_generated(self):
        s = self.reporter.compare({"findings": []}, {"findings": []})
        assert len(s.scan_before_id) > 0
        assert len(s.scan_after_id) > 0

    def test_results_format_input(self):
        before = {"results": [{"findings": [
            {"rule_id": "R1", "file": "a.py", "title": "Bug", "severity": "HIGH"},
        ]}]}
        after = {"results": [{"findings": []}]}
        s = self.reporter.compare(before, after)
        assert s.fixed_count == 1

    def test_large_diff(self):
        before_findings = [
            {"rule_id": f"R{i}", "file": f"f{i}.py", "title": f"Bug {i}", "severity": "LOW"}
            for i in range(50)
        ]
        after_findings = [
            {"rule_id": f"R{i}", "file": f"f{i}.py", "title": f"Bug {i}", "severity": "LOW"}
            for i in range(25, 75)
        ]
        s = self.reporter.compare(
            {"findings": before_findings},
            {"findings": after_findings},
        )
        assert s.fixed_count == 25
        assert s.new_count == 25
        assert s.unchanged_count == 25


# ---------- Time Series ----------

class TestTimeSeries:
    def setup_method(self):
        self.reporter = DiffReporter()

    def test_single_scan_no_comparison(self):
        r = self.reporter.compare_time_series([{"findings": []}])
        assert r.points == []

    def test_two_scans(self):
        scan1 = {"timestamp": "t1", "findings": [
            {"rule_id": "R1", "file": "a.py", "title": "Bug", "severity": "HIGH"},
        ]}
        scan2 = {"timestamp": "t2", "findings": []}
        r = self.reporter.compare_time_series([scan1, scan2])
        assert len(r.points) == 1
        assert r.points[0].fixed_count == 1

    def test_three_scans(self):
        scans = [
            {"timestamp": "t1", "findings": [
                {"rule_id": "R1", "file": "a.py", "title": "Bug", "severity": "HIGH"},
            ]},
            {"timestamp": "t2", "findings": [
                {"rule_id": "R1", "file": "a.py", "title": "Bug", "severity": "HIGH"},
                {"rule_id": "R2", "file": "b.py", "title": "New", "severity": "LOW"},
            ]},
            {"timestamp": "t3", "findings": []},
        ]
        r = self.reporter.compare_time_series(scans)
        assert len(r.points) == 2

    def test_improving_trend(self):
        scans = [
            {"timestamp": "t1", "findings": [
                {"rule_id": f"R{i}", "file": f"f{i}.py", "title": f"B{i}", "severity": "HIGH"}
                for i in range(10)
            ]},
            {"timestamp": "t2", "findings": [
                {"rule_id": f"R{i}", "file": f"f{i}.py", "title": f"B{i}", "severity": "HIGH"}
                for i in range(5)
            ]},
        ]
        r = self.reporter.compare_time_series(scans)
        assert r.overall_trend == "improving"

    def test_degrading_trend(self):
        scans = [
            {"timestamp": "t1", "findings": []},
            {"timestamp": "t2", "findings": [
                {"rule_id": f"R{i}", "file": f"f{i}.py", "title": f"B{i}", "severity": "HIGH"}
                for i in range(10)
            ]},
        ]
        r = self.reporter.compare_time_series(scans)
        assert r.overall_trend == "degrading"

    def test_stable_trend(self):
        finding = {"rule_id": "R1", "file": "a.py", "title": "Bug", "severity": "HIGH"}
        scans = [
            {"timestamp": "t1", "findings": [finding]},
            {"timestamp": "t2", "findings": [finding]},
        ]
        r = self.reporter.compare_time_series(scans)
        assert r.overall_trend == "stable"

    def test_total_counts(self):
        scans = [
            {"timestamp": "t1", "findings": [
                {"rule_id": "R1", "file": "a.py", "title": "B1", "severity": "HIGH"},
            ]},
            {"timestamp": "t2", "findings": [
                {"rule_id": "R2", "file": "b.py", "title": "B2", "severity": "LOW"},
            ]},
            {"timestamp": "t3", "findings": []},
        ]
        r = self.reporter.compare_time_series(scans)
        assert r.total_new_across_all >= 1
        assert r.total_fixed_across_all >= 1

    def test_empty_scans_list(self):
        r = self.reporter.compare_time_series([])
        assert r.points == []
        assert r.overall_trend == ""


# ---------- Filter ----------

class TestFilterEntries:
    def setup_method(self):
        self.reporter = DiffReporter()
        self.summary = DiffSummary(entries=[
            ScanDiffEntry("a", DiffStatus.NEW, Severity.HIGH, file="server.py"),
            ScanDiffEntry("b", DiffStatus.FIXED, Severity.LOW, file="client.py"),
            ScanDiffEntry("c", DiffStatus.NEW, Severity.CRITICAL, file="auth.py"),
            ScanDiffEntry("d", DiffStatus.UNCHANGED, Severity.MEDIUM, file="server.py"),
        ])

    def test_filter_by_status(self):
        result = self.reporter.filter_entries(self.summary, status=DiffStatus.NEW)
        assert len(result) == 2

    def test_filter_by_severity(self):
        result = self.reporter.filter_entries(self.summary, severity=Severity.HIGH)
        assert len(result) == 1
        assert result[0].finding_id == "a"

    def test_filter_by_file_pattern(self):
        result = self.reporter.filter_entries(self.summary, file_pattern="server")
        assert len(result) == 2

    def test_filter_combined(self):
        result = self.reporter.filter_entries(
            self.summary, status=DiffStatus.NEW, severity=Severity.HIGH,
        )
        assert len(result) == 1

    def test_filter_no_match(self):
        result = self.reporter.filter_entries(self.summary, severity="UNKNOWN")
        assert len(result) == 0


# ---------- Markdown Formatter ----------

class TestFormatMarkdown:
    def test_basic_output(self):
        s = DiffSummary(entries=[
            ScanDiffEntry("a", DiffStatus.NEW, Severity.HIGH, file="x.py", line=10, title="Bug"),
        ])
        md = format_markdown(s)
        assert "# Scan Diff Report" in md
        assert "New findings" in md
        assert "| NEW |" in md

    def test_timestamps_shown(self):
        s = DiffSummary(
            timestamp_before="2024-01-01",
            timestamp_after="2024-01-02",
        )
        md = format_markdown(s)
        assert "2024-01-01" in md
        assert "2024-01-02" in md

    def test_severity_breakdown_shown(self):
        s = DiffSummary(entries=[
            ScanDiffEntry("a", DiffStatus.NEW, Severity.CRITICAL),
        ])
        md = format_markdown(s)
        assert "CRITICAL" in md

    def test_modified_shows_arrow(self):
        s = DiffSummary(entries=[
            ScanDiffEntry(
                "a", DiffStatus.MODIFIED, Severity.CRITICAL,
                previous_severity=Severity.HIGH, file="x.py", line=1,
            ),
        ])
        md = format_markdown(s)
        assert "HIGH -> CRITICAL" in md

    def test_empty_summary(self):
        s = DiffSummary()
        md = format_markdown(s)
        assert "# Scan Diff Report" in md


# ---------- JSON Formatter ----------

class TestFormatJson:
    def test_valid_json(self):
        s = DiffSummary(entries=[
            ScanDiffEntry("a", DiffStatus.NEW, Severity.HIGH),
        ])
        text = format_json(s)
        data = json.loads(text)
        assert data["counts"]["new"] == 1

    def test_roundtrip(self):
        s = DiffSummary(
            scan_before_id="s1",
            scan_after_id="s2",
            entries=[
                ScanDiffEntry("a", DiffStatus.FIXED, Severity.LOW),
            ],
        )
        data = json.loads(format_json(s))
        assert data["scan_before_id"] == "s1"
        assert data["counts"]["fixed"] == 1


# ---------- Time Series Formatters ----------

class TestTimeSeriesFormatters:
    def test_markdown_basic(self):
        r = TimeSeriesReport(
            points=[TimeSeriesPoint("t1", 5, 2, 1, 1, -10.0)],
            overall_trend="degrading",
        )
        md = format_time_series_markdown(r)
        assert "Time Series" in md
        assert "degrading" in md

    def test_json_valid(self):
        r = TimeSeriesReport(
            points=[TimeSeriesPoint("t1")],
            overall_trend="stable",
        )
        text = format_time_series_json(r)
        data = json.loads(text)
        assert data["overall_trend"] == "stable"

    def test_empty_report_markdown(self):
        r = TimeSeriesReport()
        md = format_time_series_markdown(r)
        assert "Time Series" in md



"""Configuration Drift Detector for MCP servers.

Detects unauthorized or unexpected changes in MCP server configurations
by comparing current state against a saved baseline snapshot.
Identifies new tools, changed permissions, modified environment variables,
and new endpoints.

Zero external dependencies — pure Python stdlib.
"""

from __future__ import annotations

import hashlib
import json
import re
import time
from dataclasses import dataclass, field
from pathlib import Path
from typing import Any


# ---------- Severity ----------

class DriftSeverity:
    """Drift severity constants."""
    CRITICAL = "critical"
    HIGH = "high"
    MEDIUM = "medium"
    LOW = "low"
    INFO = "info"


# ---------- Data Models ----------

@dataclass
class DriftItem:
    """A single configuration drift finding."""

    category: str  # tools, permissions, env_vars, endpoints, config
    change_type: str  # added, removed, modified
    key: str
    old_value: str = ""
    new_value: str = ""
    severity: str = DriftSeverity.MEDIUM
    recommendation: str = ""

    def to_dict(self) -> dict[str, Any]:
        return {
            "category": self.category,
            "change_type": self.change_type,
            "key": self.key,
            "old_value": self.old_value,
            "new_value": self.new_value,
            "severity": self.severity,
            "recommendation": self.recommendation,
        }


@dataclass
class DriftReport:
    """Aggregated drift detection report."""

    baseline_path: str = ""
    current_path: str = ""
    drifts: list[DriftItem] = field(default_factory=list)
    timestamp: str = ""
    baseline_hash: str = ""
    current_hash: str = ""

    @property
    def total_drifts(self) -> int:
        return len(self.drifts)

    @property
    def critical_count(self) -> int:
        return sum(1 for d in self.drifts if d.severity == DriftSeverity.CRITICAL)

    @property
    def high_count(self) -> int:
        return sum(1 for d in self.drifts if d.severity == DriftSeverity.HIGH)

    @property
    def medium_count(self) -> int:
        return sum(1 for d in self.drifts if d.severity == DriftSeverity.MEDIUM)

    @property
    def low_count(self) -> int:
        return sum(1 for d in self.drifts if d.severity == DriftSeverity.LOW)

    @property
    def has_critical(self) -> bool:
        return self.critical_count > 0

    @property
    def severity_counts(self) -> dict[str, int]:
        return {
            "critical": self.critical_count,
            "high": self.high_count,
            "medium": self.medium_count,
            "low": self.low_count,
        }

    @property
    def exit_code(self) -> int:
        if self.critical_count > 0:
            return 2
        if self.high_count > 0:
            return 1
        return 0

    def to_dict(self) -> dict[str, Any]:
        return {
            "baseline_path": self.baseline_path,
            "current_path": self.current_path,
            "total_drifts": self.total_drifts,
            "severity_counts": self.severity_counts,
            "exit_code": self.exit_code,
            "timestamp": self.timestamp,
            "baseline_hash": self.baseline_hash,
            "current_hash": self.current_hash,
            "drifts": [d.to_dict() for d in self.drifts],
        }


@dataclass
class ConfigSnapshot:
    """Snapshot of an MCP server configuration."""

    path: str = ""
    timestamp: str = ""
    config_hash: str = ""
    tools: dict[str, Any] = field(default_factory=dict)
    permissions: dict[str, Any] = field(default_factory=dict)
    env_vars: dict[str, str] = field(default_factory=dict)
    endpoints: list[str] = field(default_factory=list)
    raw_config: dict[str, Any] = field(default_factory=dict)

    def to_dict(self) -> dict[str, Any]:
        return {
            "path": self.path,
            "timestamp": self.timestamp,
            "config_hash": self.config_hash,
            "tools": self.tools,
            "permissions": self.permissions,
            "env_vars": self.env_vars,
            "endpoints": self.endpoints,
            "raw_config": self.raw_config,
        }

    @classmethod
    def from_dict(cls, data: dict[str, Any]) -> ConfigSnapshot:
        """Create a ConfigSnapshot from a dict."""
        return cls(
            path=data.get("path", ""),
            timestamp=data.get("timestamp", ""),
            config_hash=data.get("config_hash", ""),
            tools=data.get("tools", {}),
            permissions=data.get("permissions", {}),
            env_vars=data.get("env_vars", {}),
            endpoints=data.get("endpoints", []),
            raw_config=data.get("raw_config", {}),
        )


# ---------- Config Extraction ----------

def _compute_hash(data: dict[str, Any]) -> str:
    """Compute SHA-256 hash of config data."""
    raw = json.dumps(data, sort_keys=True, ensure_ascii=False)
    return hashlib.sha256(raw.encode("utf-8")).hexdigest()[:16]


def _extract_tools_from_config(config: dict[str, Any]) -> dict[str, Any]:
    """Extract tool definitions from MCP config JSON."""
    tools: dict[str, Any] = {}
    # Claude Desktop format: mcpServers -> server -> tools
    for server_name, server_config in config.get("mcpServers", {}).items():
        if isinstance(server_config, dict):
            server_tools = server_config.get("tools", {})
            if isinstance(server_tools, dict):
                for tool_name, tool_def in server_tools.items():
                    tools[f"{server_name}.{tool_name}"] = tool_def
            elif isinstance(server_tools, list):
                for tool in server_tools:
                    if isinstance(tool, str):
                        tools[f"{server_name}.{tool}"] = {}
                    elif isinstance(tool, dict) and "name" in tool:
                        tools[f"{server_name}.{tool['name']}"] = tool
    # Generic tools key at top level
    top_tools = config.get("tools", {})
    if isinstance(top_tools, dict):
        for name, defn in top_tools.items():
            tools[name] = defn
    elif isinstance(top_tools, list):
        for tool in top_tools:
            if isinstance(tool, str):
                tools[tool] = {}
            elif isinstance(tool, dict) and "name" in tool:
                tools[tool["name"]] = tool
    return tools


def _extract_permissions(config: dict[str, Any]) -> dict[str, Any]:
    """Extract permission settings from MCP config."""
    perms: dict[str, Any] = {}
    # Check mcpServers for per-server permissions
    for server_name, server_config in config.get("mcpServers", {}).items():
        if isinstance(server_config, dict):
            if "permissions" in server_config:
                perms[server_name] = server_config["permissions"]
            if "allowedDirectories" in server_config:
                perms[f"{server_name}.allowedDirectories"] = server_config["allowedDirectories"]
            if "allowedCommands" in server_config:
                perms[f"{server_name}.allowedCommands"] = server_config["allowedCommands"]
    # Top-level permissions
    if "permissions" in config:
        perms["_global"] = config["permissions"]
    if "allowedDirectories" in config:
        perms["_global.allowedDirectories"] = config["allowedDirectories"]
    return perms


def _extract_env_vars(config: dict[str, Any]) -> dict[str, str]:
    """Extract environment variable settings from MCP config."""
    env_vars: dict[str, str] = {}
    for server_name, server_config in config.get("mcpServers", {}).items():
        if isinstance(server_config, dict):
            env = server_config.get("env", {})
            if isinstance(env, dict):
                for k, v in env.items():
                    env_vars[f"{server_name}.{k}"] = str(v)
    # Top-level env
    top_env = config.get("env", {})
    if isinstance(top_env, dict):
        for k, v in top_env.items():
            env_vars[k] = str(v)
    return env_vars


def _extract_endpoints(config: dict[str, Any]) -> list[str]:
    """Extract server endpoints/commands from MCP config."""
    endpoints: list[str] = []
    for server_name, server_config in config.get("mcpServers", {}).items():
        if isinstance(server_config, dict):
            cmd = server_config.get("command", "")
            args = server_config.get("args", [])
            if cmd:
                full = f"{server_name}: {cmd}"
                if args and isinstance(args, list):
                    full += " " + " ".join(str(a) for a in args)
                endpoints.append(full)
    return sorted(endpoints)


# ---------- Snapshot Creation ----------

def create_snapshot(path: Path) -> ConfigSnapshot:
    """Create a configuration snapshot from a file or directory.

    Args:
        path: Path to MCP config file (JSON) or directory containing one.

    Returns:
        ConfigSnapshot with extracted configuration data.
    """
    config_file = _find_config_file(path)
    if config_file is None:
        return ConfigSnapshot(
            path=str(path),
            timestamp=_now_iso(),
        )

    try:
        content = config_file.read_text(encoding="utf-8", errors="replace")
        config = json.loads(content)
    except (json.JSONDecodeError, OSError):
        return ConfigSnapshot(
            path=str(config_file),
            timestamp=_now_iso(),
        )

    tools = _extract_tools_from_config(config)
    permissions = _extract_permissions(config)
    env_vars = _extract_env_vars(config)
    endpoints = _extract_endpoints(config)

    snapshot = ConfigSnapshot(
        path=str(config_file),
        timestamp=_now_iso(),
        tools=tools,
        permissions=permissions,
        env_vars=env_vars,
        endpoints=endpoints,
        raw_config=config,
    )
    snapshot.config_hash = _compute_hash(config)
    return snapshot


def create_snapshot_from_dict(config: dict[str, Any], path: str = "") -> ConfigSnapshot:
    """Create a configuration snapshot from a dict.

    Args:
        config: MCP config dict.
        path: Optional path label.

    Returns:
        ConfigSnapshot.
    """
    tools = _extract_tools_from_config(config)
    permissions = _extract_permissions(config)
    env_vars = _extract_env_vars(config)
    endpoints = _extract_endpoints(config)

    snapshot = ConfigSnapshot(
        path=path,
        timestamp=_now_iso(),
        tools=tools,
        permissions=permissions,
        env_vars=env_vars,
        endpoints=endpoints,
        raw_config=config,
    )
    snapshot.config_hash = _compute_hash(config)
    return snapshot


def save_snapshot(snapshot: ConfigSnapshot, output_path: Path) -> None:
    """Save a snapshot to a JSON file."""
    output_path.write_text(
        json.dumps(snapshot.to_dict(), indent=2, ensure_ascii=False),
        encoding="utf-8",
    )


def load_snapshot(path: Path) -> ConfigSnapshot:
    """Load a snapshot from a JSON file."""
    content = path.read_text(encoding="utf-8", errors="replace")
    data = json.loads(content)
    return ConfigSnapshot.from_dict(data)


# ---------- Drift Detection ----------

def _severity_for_category(category: str, change_type: str) -> str:
    """Determine severity based on drift category and change type."""
    if category == "tools" and change_type == "added":
        return DriftSeverity.HIGH
    if category == "tools" and change_type == "removed":
        return DriftSeverity.MEDIUM
    if category == "tools" and change_type == "modified":
        return DriftSeverity.HIGH
    if category == "permissions":
        return DriftSeverity.CRITICAL
    if category == "env_vars" and change_type == "added":
        return DriftSeverity.MEDIUM
    if category == "env_vars" and change_type == "modified":
        return DriftSeverity.HIGH
    if category == "env_vars" and change_type == "removed":
        return DriftSeverity.LOW
    if category == "endpoints" and change_type == "added":
        return DriftSeverity.HIGH
    if category == "endpoints" and change_type == "removed":
        return DriftSeverity.MEDIUM
    return DriftSeverity.MEDIUM


def _recommendation_for_drift(item: DriftItem) -> str:
    """Generate a recommendation for a drift item."""
    recs = {
        ("tools", "added"): "Review the new tool for security implications before deployment",
        ("tools", "removed"): "Verify tool removal was intentional and documented",
        ("tools", "modified"): "Review tool definition changes for security regressions",
        ("permissions", "added"): "Audit new permissions — restrict to minimum required access",
        ("permissions", "removed"): "Verify permission removal does not break functionality",
        ("permissions", "modified"): "Review permission changes — ensure no privilege escalation",
        ("env_vars", "added"): "Check if new environment variable contains secrets or sensitive data",
        ("env_vars", "removed"): "Verify env var removal was intentional",
        ("env_vars", "modified"): "Review env var change — check for leaked secrets or misconfig",
        ("endpoints", "added"): "Audit new endpoint for security — validate command and arguments",
        ("endpoints", "removed"): "Verify endpoint removal was planned",
    }
    return recs.get((item.category, item.change_type), "Review this configuration change")


def _compare_dicts(
    baseline: dict[str, Any],
    current: dict[str, Any],
    category: str,
) -> list[DriftItem]:
    """Compare two dicts and return drift items."""
    drifts: list[DriftItem] = []
    baseline_keys = set(baseline.keys())
    current_keys = set(current.keys())

    # Added keys
    for key in sorted(current_keys - baseline_keys):
        item = DriftItem(
            category=category,
            change_type="added",
            key=key,
            new_value=json.dumps(current[key], ensure_ascii=False)
            if not isinstance(current[key], str) else current[key],
            severity=_severity_for_category(category, "added"),
        )
        item.recommendation = _recommendation_for_drift(item)
        drifts.append(item)

    # Removed keys
    for key in sorted(baseline_keys - current_keys):
        item = DriftItem(
            category=category,
            change_type="removed",
            key=key,
            old_value=json.dumps(baseline[key], ensure_ascii=False)
            if not isinstance(baseline[key], str) else baseline[key],
            severity=_severity_for_category(category, "removed"),
        )
        item.recommendation = _recommendation_for_drift(item)
        drifts.append(item)

    # Modified keys
    for key in sorted(baseline_keys & current_keys):
        old_val = baseline[key]
        new_val = current[key]
        if old_val != new_val:
            item = DriftItem(
                category=category,
                change_type="modified",
                key=key,
                old_value=json.dumps(old_val, ensure_ascii=False)
                if not isinstance(old_val, str) else old_val,
                new_value=json.dumps(new_val, ensure_ascii=False)
                if not isinstance(new_val, str) else new_val,
                severity=_severity_for_category(category, "modified"),
            )
            item.recommendation = _recommendation_for_drift(item)
            drifts.append(item)

    return drifts


def _compare_lists(
    baseline: list[str],
    current: list[str],
    category: str,
) -> list[DriftItem]:
    """Compare two lists and return drift items."""
    drifts: list[DriftItem] = []
    baseline_set = set(baseline)
    current_set = set(current)

    for item_str in sorted(current_set - baseline_set):
        item = DriftItem(
            category=category,
            change_type="added",
            key=item_str,
            new_value=item_str,
            severity=_severity_for_category(category, "added"),
        )
        item.recommendation = _recommendation_for_drift(item)
        drifts.append(item)

    for item_str in sorted(baseline_set - current_set):
        item = DriftItem(
            category=category,
            change_type="removed",
            key=item_str,
            old_value=item_str,
            severity=_severity_for_category(category, "removed"),
        )
        item.recommendation = _recommendation_for_drift(item)
        drifts.append(item)

    return drifts


def check_drift(
    baseline: ConfigSnapshot,
    current: ConfigSnapshot,
) -> DriftReport:
    """Compare two snapshots and detect configuration drift.

    Args:
        baseline: The baseline (known-good) snapshot.
        current: The current snapshot to compare.

    Returns:
        DriftReport with all detected drifts.
    """
    report = DriftReport(
        baseline_path=baseline.path,
        current_path=current.path,
        timestamp=_now_iso(),
        baseline_hash=baseline.config_hash,
        current_hash=current.config_hash,
    )

    # Compare tools
    report.drifts.extend(_compare_dicts(baseline.tools, current.tools, "tools"))
    # Compare permissions
    report.drifts.extend(_compare_dicts(baseline.permissions, current.permissions, "permissions"))
    # Compare env vars
    report.drifts.extend(_compare_dicts(baseline.env_vars, current.env_vars, "env_vars"))
    # Compare endpoints
    report.drifts.extend(_compare_lists(baseline.endpoints, current.endpoints, "endpoints"))

    return report


def check_drift_from_file(
    baseline_path: Path,
    current_path: Path,
) -> DriftReport:
    """Compare a baseline snapshot file against a current config.

    Args:
        baseline_path: Path to saved baseline snapshot JSON.
        current_path: Path to current MCP config file or directory.

    Returns:
        DriftReport.
    """
    baseline = load_snapshot(baseline_path)
    current = create_snapshot(current_path)
    return check_drift(baseline, current)


# ---------- Helpers ----------

def _now_iso() -> str:
    """Return current timestamp as ISO 8601 string."""
    return time.strftime("%Y-%m-%dT%H:%M:%SZ", time.gmtime())


def _find_config_file(path: Path) -> Path | None:
    """Find MCP config file from a path.

    Looks for common MCP config filenames.
    """
    if path.is_file():
        return path

    if path.is_dir():
        candidates = [
            "claude_desktop_config.json",
            "mcp_config.json",
            "mcp.json",
            ".mcp.json",
            "config.json",
        ]
        for name in candidates:
            candidate = path / name
            if candidate.exists():
                return candidate
    return None


# ---------- Formatters ----------

def format_drift_text(report: DriftReport) -> str:
    """Format a DriftReport as human-readable text."""
    lines: list[str] = []
    lines.append(f"\n  Configuration Drift Report")
    lines.append(f"  Baseline: {report.baseline_path}")
    lines.append(f"  Current:  {report.current_path}")
    lines.append(f"  Total drifts: {report.total_drifts}")
    lines.append(f"  Severity: critical={report.critical_count} "
                 f"high={report.high_count} "
                 f"medium={report.medium_count} "
                 f"low={report.low_count}")
    lines.append("")

    if not report.drifts:
        lines.append("  No configuration drift detected.")
        return "\n".join(lines)

    severity_icons = {
        DriftSeverity.CRITICAL: "[!!]",
        DriftSeverity.HIGH: "[!]",
        DriftSeverity.MEDIUM: "[-]",
        DriftSeverity.LOW: "[.]",
        DriftSeverity.INFO: "[i]",
    }

    # Group by category
    categories = sorted({d.category for d in report.drifts})
    for cat in categories:
        cat_drifts = [d for d in report.drifts if d.category == cat]
        lines.append(f"  {cat.upper()} ({len(cat_drifts)} changes):")
        for d in cat_drifts:
            icon = severity_icons.get(d.severity, "[-]")
            lines.append(f"    {icon} {d.change_type}: {d.key}")
            if d.old_value:
                lines.append(f"        old: {d.old_value}")
            if d.new_value:
                lines.append(f"        new: {d.new_value}")
            if d.recommendation:
                lines.append(f"        rec: {d.recommendation}")
        lines.append("")

    return "\n".join(lines)


def format_drift_json(report: DriftReport) -> str:
    """Format a DriftReport as JSON."""
    return json.dumps(report.to_dict(), indent=2, ensure_ascii=False)



"""Tests for Finding Annotator."""

from __future__ import annotations

import json
from pathlib import Path
from typing import Any

import pytest

from mcpaudit.annotator import (
    Annotation,
    AnnotationStore,
    DEFAULT_ANNOTATIONS_FILE,
    EXCLUDE_STATUSES,
    VALID_STATUSES,
    annotate_finding,
    bulk_annotate,
    filter_annotated,
    filter_by_status,
    format_annotations_json,
    format_annotations_text,
    load_and_annotate,
    load_annotations,
    make_finding_id,
    make_finding_id_from_dict,
    remove_annotation,
    save_annotations,
)


# ---------------------------------------------------------------------------
# Helpers
# ---------------------------------------------------------------------------

def _sample_findings() -> list[dict[str, Any]]:
    return [
        {"rule_id": "CMD-001", "file": "server.py", "line": 10,
         "severity": "HIGH", "title": "Shell injection"},
        {"rule_id": "AUTH-001", "file": "server.py", "line": 20,
         "severity": "MEDIUM", "title": "Missing auth"},
        {"rule_id": "RATE-001", "file": "api.py", "line": 5,
         "severity": "LOW", "title": "No rate limit"},
    ]


# ---------------------------------------------------------------------------
# Constants
# ---------------------------------------------------------------------------

class TestConstants:
    def test_valid_statuses(self):
        expected = {
            "false_positive", "accepted_risk", "todo",
            "in_progress", "wont_fix",
        }
        assert VALID_STATUSES == expected

    def test_exclude_statuses(self):
        assert "false_positive" in EXCLUDE_STATUSES
        assert "accepted_risk" in EXCLUDE_STATUSES
        assert "wont_fix" in EXCLUDE_STATUSES
        assert "todo" not in EXCLUDE_STATUSES
        assert "in_progress" not in EXCLUDE_STATUSES


# ---------------------------------------------------------------------------
# make_finding_id
# ---------------------------------------------------------------------------

class TestMakeFindingId:
    def test_basic(self):
        fid = make_finding_id("CMD-001", "server.py", 42)
        assert fid == "CMD-001:server.py:42"

    def test_with_path(self):
        fid = make_finding_id("CMD-001", "/path/to/server.py", 10)
        assert fid == "CMD-001:/path/to/server.py:10"

    def test_from_dict(self):
        finding = {
            "rule_id": "CMD-001",
            "file": "server.py",
            "line": 42,
        }
        fid = make_finding_id_from_dict(finding)
        assert fid == "CMD-001:server.py:42"

    def test_from_dict_missing_keys(self):
        fid = make_finding_id_from_dict({})
        assert fid == "::0"

    def test_from_dict_partial(self):
        fid = make_finding_id_from_dict({"rule_id": "CMD-001"})
        assert fid == "CMD-001::0"


# ---------------------------------------------------------------------------
# Annotation
# ---------------------------------------------------------------------------

class TestAnnotation:
    def test_create(self):
        ann = Annotation(
            finding_id="CMD-001:server.py:42",
            status="false_positive",
            note="Checked manually",
            author="carlos",
        )
        assert ann.finding_id == "CMD-001:server.py:42"
        assert ann.status == "false_positive"
        assert ann.note == "Checked manually"
        assert ann.created_at != ""

    def test_validate_valid(self):
        ann = Annotation(
            finding_id="CMD-001:a.py:1",
            status="false_positive",
        )
        assert ann.validate() == []

    def test_validate_empty_finding_id(self):
        ann = Annotation(finding_id="", status="false_positive")
        errors = ann.validate()
        assert any("finding_id" in e for e in errors)

    def test_validate_empty_status(self):
        ann = Annotation(finding_id="CMD-001:a.py:1", status="")
        errors = ann.validate()
        assert any("status" in e for e in errors)

    def test_validate_invalid_status(self):
        ann = Annotation(
            finding_id="CMD-001:a.py:1",
            status="invalid_status",
        )
        errors = ann.validate()
        assert any("Invalid status" in e for e in errors)

    def test_all_valid_statuses(self):
        for status in VALID_STATUSES:
            ann = Annotation(
                finding_id="CMD-001:a.py:1",
                status=status,
            )
            assert ann.validate() == []

    def test_to_dict(self):
        ann = Annotation(
            finding_id="CMD-001:a.py:1",
            status="todo",
            note="Fix later",
            author="test",
        )
        d = ann.to_dict()
        assert d["finding_id"] == "CMD-001:a.py:1"
        assert d["status"] == "todo"
        assert d["note"] == "Fix later"
        assert d["author"] == "test"
        assert "created_at" in d
        assert "updated_at" in d

    def test_from_dict(self):
        data = {
            "finding_id": "CMD-001:a.py:1",
            "status": "false_positive",
            "note": "OK",
            "author": "test",
            "created_at": "2024-01-01T00:00:00Z",
            "updated_at": "2024-01-01T00:00:00Z",
        }
        ann = Annotation.from_dict(data)
        assert ann.finding_id == "CMD-001:a.py:1"
        assert ann.status == "false_positive"
        assert ann.created_at == "2024-01-01T00:00:00Z"

    def test_from_dict_missing_keys(self):
        ann = Annotation.from_dict({})
        assert ann.finding_id == ""
        assert ann.status == ""

    def test_auto_timestamps(self):
        ann = Annotation(
            finding_id="CMD-001:a.py:1",
            status="todo",
        )
        assert ann.created_at != ""
        assert ann.updated_at != ""
        assert ann.created_at == ann.updated_at

    def test_preserve_timestamps(self):
        ann = Annotation(
            finding_id="CMD-001:a.py:1",
            status="todo",
            created_at="2024-01-01T00:00:00Z",
            updated_at="2024-01-02T00:00:00Z",
        )
        assert ann.created_at == "2024-01-01T00:00:00Z"
        assert ann.updated_at == "2024-01-02T00:00:00Z"


# ---------------------------------------------------------------------------
# AnnotationStore
# ---------------------------------------------------------------------------

class TestAnnotationStore:
    def test_empty_store(self):
        store = AnnotationStore()
        assert store.count == 0
        assert store.list_all() == []

    def test_add_annotation(self):
        store = AnnotationStore()
        ann = Annotation(
            finding_id="CMD-001:a.py:1",
            status="todo",
        )
        store.annotations[ann.finding_id] = ann
        assert store.count == 1

    def test_get(self):
        store = AnnotationStore()
        ann = Annotation(
            finding_id="CMD-001:a.py:1",
            status="todo",
        )
        store.annotations[ann.finding_id] = ann
        got = store.get("CMD-001:a.py:1")
        assert got is not None
        assert got.status == "todo"

    def test_get_missing(self):
        store = AnnotationStore()
        assert store.get("missing") is None

    def test_has(self):
        store = AnnotationStore()
        ann = Annotation(
            finding_id="CMD-001:a.py:1",
            status="todo",
        )
        store.annotations[ann.finding_id] = ann
        assert store.has("CMD-001:a.py:1") is True
        assert store.has("missing") is False

    def test_is_excluded_false_positive(self):
        store = AnnotationStore()
        ann = Annotation(
            finding_id="CMD-001:a.py:1",
            status="false_positive",
        )
        store.annotations[ann.finding_id] = ann
        assert store.is_excluded("CMD-001:a.py:1") is True

    def test_is_excluded_todo(self):
        store = AnnotationStore()
        ann = Annotation(
            finding_id="CMD-001:a.py:1",
            status="todo",
        )
        store.annotations[ann.finding_id] = ann
        assert store.is_excluded("CMD-001:a.py:1") is False

    def test_is_excluded_missing(self):
        store = AnnotationStore()
        assert store.is_excluded("missing") is False

    def test_filter_by_status(self):
        store = AnnotationStore()
        store.annotations["a"] = Annotation(
            finding_id="a", status="todo",
        )
        store.annotations["b"] = Annotation(
            finding_id="b", status="false_positive",
        )
        store.annotations["c"] = Annotation(
            finding_id="c", status="todo",
        )
        todos = store.filter_by_status("todo")
        assert len(todos) == 2

    def test_list_all_sorted(self):
        store = AnnotationStore()
        store.annotations["z"] = Annotation(
            finding_id="z", status="todo",
        )
        store.annotations["a"] = Annotation(
            finding_id="a", status="todo",
        )
        result = store.list_all()
        assert result[0].finding_id == "a"
        assert result[1].finding_id == "z"

    def test_to_dict(self):
        store = AnnotationStore()
        store.annotations["a"] = Annotation(
            finding_id="a", status="todo",
        )
        d = store.to_dict()
        assert "a" in d
        assert d["a"]["status"] == "todo"

    def test_from_dict(self):
        data = {
            "CMD-001:a.py:1": {
                "finding_id": "CMD-001:a.py:1",
                "status": "false_positive",
                "note": "OK",
            },
        }
        store = AnnotationStore.from_dict(data)
        assert store.count == 1
        assert store.get("CMD-001:a.py:1") is not None

    def test_from_dict_empty(self):
        store = AnnotationStore.from_dict({})
        assert store.count == 0

    def test_from_dict_invalid_entries(self):
        data = {
            "a": "not a dict",
            "b": {"status": "todo"},
        }
        store = AnnotationStore.from_dict(data)
        # "a" is invalid, "b" has empty finding_id set from key
        assert store.count >= 1

    def test_stats(self):
        store = AnnotationStore()
        store.annotations["a"] = Annotation(
            finding_id="a", status="todo",
        )
        store.annotations["b"] = Annotation(
            finding_id="b", status="todo",
        )
        store.annotations["c"] = Annotation(
            finding_id="c", status="false_positive",
        )
        stats = store.stats()
        assert stats["todo"] == 2
        assert stats["false_positive"] == 1


# ---------------------------------------------------------------------------
# annotate_finding
# ---------------------------------------------------------------------------

class TestAnnotateFinding:
    def test_basic(self):
        ann, errors = annotate_finding(
            "CMD-001:a.py:1", "false_positive",
        )
        assert errors == []
        assert ann.status == "false_positive"

    def test_with_note_and_author(self):
        ann, errors = annotate_finding(
            "CMD-001:a.py:1", "accepted_risk",
            note="Checked", author="carlos",
        )
        assert errors == []
        assert ann.note == "Checked"
        assert ann.author == "carlos"

    def test_invalid_status(self):
        ann, errors = annotate_finding(
            "CMD-001:a.py:1", "invalid",
        )
        assert len(errors) > 0

    def test_empty_finding_id(self):
        ann, errors = annotate_finding("", "todo")
        assert len(errors) > 0

    def test_updates_store(self):
        store = AnnotationStore()
        ann, errors = annotate_finding(
            "CMD-001:a.py:1", "todo",
            store=store,
        )
        assert errors == []
        assert store.count == 1
        assert store.get("CMD-001:a.py:1") is not None

    def test_updates_existing(self):
        store = AnnotationStore()
        annotate_finding(
            "CMD-001:a.py:1", "todo",
            note="First", store=store,
        )
        annotate_finding(
            "CMD-001:a.py:1", "false_positive",
            note="Updated", store=store,
        )
        assert store.count == 1
        ann = store.get("CMD-001:a.py:1")
        assert ann.status == "false_positive"
        assert ann.note == "Updated"

    def test_preserves_created_at_on_update(self):
        store = AnnotationStore()
        ann1, _ = annotate_finding(
            "CMD-001:a.py:1", "todo",
            store=store,
        )
        original_created = ann1.created_at
        ann2, _ = annotate_finding(
            "CMD-001:a.py:1", "false_positive",
            store=store,
        )
        assert ann2.created_at == original_created


# ---------------------------------------------------------------------------
# bulk_annotate
# ---------------------------------------------------------------------------

class TestBulkAnnotate:
    def test_basic(self):
        store = AnnotationStore()
        results, errors = bulk_annotate(
            ["a", "b", "c"], "todo",
            store=store,
        )
        assert errors == []
        assert len(results) == 3
        assert store.count == 3

    def test_invalid_status(self):
        results, errors = bulk_annotate(
            ["a", "b"], "invalid",
        )
        assert len(errors) > 0
        assert len(results) == 0

    def test_with_note(self):
        store = AnnotationStore()
        results, errors = bulk_annotate(
            ["a", "b"], "accepted_risk",
            note="Bulk accepted", author="bot",
            store=store,
        )
        assert errors == []
        for ann in store.list_all():
            assert ann.note == "Bulk accepted"
            assert ann.author == "bot"

    def test_empty_list(self):
        store = AnnotationStore()
        results, errors = bulk_annotate([], "todo", store=store)
        assert results == []
        assert errors == []


# ---------------------------------------------------------------------------
# remove_annotation
# ---------------------------------------------------------------------------

class TestRemoveAnnotation:
    def test_remove_existing(self):
        store = AnnotationStore()
        annotate_finding("a", "todo", store=store)
        assert remove_annotation("a", store) is True
        assert store.count == 0

    def test_remove_missing(self):
        store = AnnotationStore()
        assert remove_annotation("missing", store) is False


# ---------------------------------------------------------------------------
# filter_annotated
# ---------------------------------------------------------------------------

class TestFilterAnnotated:
    def test_no_annotations(self):
        findings = _sample_findings()
        result = filter_annotated(findings, {})
        assert len(result) == 3

    def test_filter_false_positive(self):
        findings = _sample_findings()
        store = AnnotationStore()
        fid = make_finding_id_from_dict(findings[0])
        annotate_finding(fid, "false_positive", store=store)
        result = filter_annotated(findings, store)
        assert len(result) == 2

    def test_filter_accepted_risk(self):
        findings = _sample_findings()
        store = AnnotationStore()
        fid = make_finding_id_from_dict(findings[1])
        annotate_finding(fid, "accepted_risk", store=store)
        result = filter_annotated(findings, store)
        assert len(result) == 2

    def test_filter_wont_fix(self):
        findings = _sample_findings()
        store = AnnotationStore()
        fid = make_finding_id_from_dict(findings[0])
        annotate_finding(fid, "wont_fix", store=store)
        result = filter_annotated(findings, store)
        assert len(result) == 2

    def test_todo_not_filtered(self):
        findings = _sample_findings()
        store = AnnotationStore()
        fid = make_finding_id_from_dict(findings[0])
        annotate_finding(fid, "todo", store=store)
        result = filter_annotated(findings, store)
        assert len(result) == 3

    def test_in_progress_not_filtered(self):
        findings = _sample_findings()
        store = AnnotationStore()
        fid = make_finding_id_from_dict(findings[0])
        annotate_finding(fid, "in_progress", store=store)
        result = filter_annotated(findings, store)
        assert len(result) == 3

    def test_custom_exclude_statuses(self):
        findings = _sample_findings()
        store = AnnotationStore()
        fid = make_finding_id_from_dict(findings[0])
        annotate_finding(fid, "todo", store=store)
        result = filter_annotated(
            findings, store,
            exclude_statuses=frozenset({"todo"}),
        )
        assert len(result) == 2

    def test_filter_with_dict(self):
        findings = _sample_findings()
        fid = make_finding_id_from_dict(findings[0])
        annotations = {
            fid: {
                "finding_id": fid,
                "status": "false_positive",
            },
        }
        result = filter_annotated(findings, annotations)
        assert len(result) == 2

    def test_filter_all(self):
        findings = _sample_findings()
        store = AnnotationStore()
        for f in findings:
            fid = make_finding_id_from_dict(f)
            annotate_finding(fid, "false_positive", store=store)
        result = filter_annotated(findings, store)
        assert len(result) == 0


# ---------------------------------------------------------------------------
# filter_by_status
# ---------------------------------------------------------------------------

class TestFilterByStatus:
    def test_basic(self):
        findings = _sample_findings()
        store = AnnotationStore()
        fid0 = make_finding_id_from_dict(findings[0])
        fid1 = make_finding_id_from_dict(findings[1])
        annotate_finding(fid0, "todo", store=store)
        annotate_finding(fid1, "false_positive", store=store)

        todos = filter_by_status(findings, store, "todo")
        assert len(todos) == 1

        fps = filter_by_status(findings, store, "false_positive")
        assert len(fps) == 1

    def test_no_matches(self):
        findings = _sample_findings()
        store = AnnotationStore()
        result = filter_by_status(findings, store, "todo")
        assert len(result) == 0

    def test_with_dict(self):
        findings = _sample_findings()
        fid = make_finding_id_from_dict(findings[0])
        annotations = {
            fid: {"finding_id": fid, "status": "todo"},
        }
        result = filter_by_status(findings, annotations, "todo")
        assert len(result) == 1


# ---------------------------------------------------------------------------
# Persistence
# ---------------------------------------------------------------------------

class TestPersistence:
    def test_save_and_load(self, tmp_path):
        path = tmp_path / "annotations.json"
        store = AnnotationStore()
        annotate_finding("a", "todo", note="Test", store=store)

        save_annotations(store, path)
        assert path.exists()

        loaded = load_annotations(path)
        assert loaded.count == 1
        ann = loaded.get("a")
        assert ann is not None
        assert ann.status == "todo"
        assert ann.note == "Test"

    def test_load_nonexistent(self, tmp_path):
        path = tmp_path / "missing.json"
        store = load_annotations(path)
        assert store.count == 0

    def test_load_invalid_json(self, tmp_path):
        path = tmp_path / "bad.json"
        path.write_text("not valid json", encoding="utf-8")
        store = load_annotations(path)
        assert store.count == 0

    def test_load_non_dict(self, tmp_path):
        path = tmp_path / "array.json"
        path.write_text("[1, 2, 3]", encoding="utf-8")
        store = load_annotations(path)
        assert store.count == 0

    def test_save_creates_file(self, tmp_path):
        path = tmp_path / "new.json"
        store = AnnotationStore()
        save_annotations(store, path)
        assert path.exists()

    def test_roundtrip_preserves_data(self, tmp_path):
        path = tmp_path / "rt.json"
        store = AnnotationStore()
        annotate_finding(
            "CMD-001:a.py:1", "false_positive",
            note="Checked", author="carlos",
            store=store,
        )
        save_annotations(store, path)
        loaded = load_annotations(path)
        ann = loaded.get("CMD-001:a.py:1")
        assert ann is not None
        assert ann.status == "false_positive"
        assert ann.note == "Checked"
        assert ann.author == "carlos"


# ---------------------------------------------------------------------------
# load_and_annotate
# ---------------------------------------------------------------------------

class TestLoadAndAnnotate:
    def test_basic(self, tmp_path):
        path = tmp_path / "ann.json"
        ann, errors = load_and_annotate(
            "CMD-001:a.py:1", "todo",
            note="Fix this", path=path,
        )
        assert errors == []
        assert ann.status == "todo"
        assert path.exists()

        # Verify it was saved
        loaded = load_annotations(path)
        assert loaded.count == 1

    def test_update_existing(self, tmp_path):
        path = tmp_path / "ann.json"
        load_and_annotate(
            "CMD-001:a.py:1", "todo", path=path,
        )
        load_and_annotate(
            "CMD-001:a.py:1", "false_positive",
            note="Verified", path=path,
        )
        loaded = load_annotations(path)
        assert loaded.count == 1
        ann = loaded.get("CMD-001:a.py:1")
        assert ann.status == "false_positive"

    def test_invalid_does_not_save(self, tmp_path):
        path = tmp_path / "ann.json"
        ann, errors = load_and_annotate(
            "CMD-001:a.py:1", "invalid",
            path=path,
        )
        assert len(errors) > 0
        assert not path.exists()


# ---------------------------------------------------------------------------
# Formatting
# ---------------------------------------------------------------------------

class TestFormatAnnotationsText:
    def test_empty(self):
        store = AnnotationStore()
        text = format_annotations_text(store)
        assert "No annotations" in text

    def test_with_annotations(self):
        store = AnnotationStore()
        annotate_finding(
            "CMD-001:a.py:1", "todo",
            note="Fix this", author="carlos",
            store=store,
        )
        text = format_annotations_text(store)
        assert "CMD-001:a.py:1" in text
        assert "todo" in text
        assert "Fix this" in text
        assert "carlos" in text

    def test_stats_shown(self):
        store = AnnotationStore()
        annotate_finding("a", "todo", store=store)
        annotate_finding("b", "false_positive", store=store)
        text = format_annotations_text(store)
        assert "todo: 1" in text
        assert "false_positive: 1" in text


class TestFormatAnnotationsJson:
    def test_empty(self):
        store = AnnotationStore()
        j = format_annotations_json(store)
        data = json.loads(j)
        assert data["count"] == 0
        assert data["annotations"] == {}

    def test_with_annotations(self):
        store = AnnotationStore()
        annotate_finding("a", "todo", store=store)
        j = format_annotations_json(store)
        data = json.loads(j)
        assert data["count"] == 1
        assert "a" in data["annotations"]

    def test_stats_included(self):
        store = AnnotationStore()
        annotate_finding("a", "todo", store=store)
        annotate_finding("b", "false_positive", store=store)
        j = format_annotations_json(store)
        data = json.loads(j)
        assert data["stats"]["todo"] == 1
        assert data["stats"]["false_positive"] == 1



"""MCP Server Fingerprinting -- Detect framework and apply specific rules.

Analyzes source code to determine which MCP framework a server uses
(FastMCP, mcp-python-sdk, @modelcontextprotocol/sdk, custom) and
provides framework-specific security recommendations and rules.

Zero external dependencies. All detection is regex-based.
"""

from __future__ import annotations

import re
from dataclasses import dataclass, field
from pathlib import Path
from typing import Any


# ---------------------------------------------------------------------------
# Data Models
# ---------------------------------------------------------------------------

@dataclass
class KnownVulnerability:
    """A known vulnerability for a specific MCP framework.

    Attributes:
        id: Vulnerability identifier.
        title: Short title.
        description: Detailed description.
        severity: Severity string (CRITICAL, HIGH, MEDIUM, LOW).
        cwe: CWE reference.
        affected_versions: Version range affected.
        fix: Remediation guidance.
    """

    id: str
    title: str
    description: str
    severity: str = "MEDIUM"
    cwe: str = ""
    affected_versions: str = ""
    fix: str = ""

    def to_dict(self) -> dict[str, Any]:
        return {
            "id": self.id,
            "title": self.title,
            "description": self.description,
            "severity": self.severity,
            "cwe": self.cwe,
            "affected_versions": self.affected_versions,
            "fix": self.fix,
        }


@dataclass
class FrameworkInfo:
    """Information about a detected MCP framework.

    Attributes:
        name: Framework name (e.g. 'FastMCP', 'mcp-python-sdk').
        version_hint: Detected or inferred version hint.
        known_vulnerabilities: Known security issues for this framework.
        recommended_rules: Rule IDs recommended for this framework.
        confidence: Detection confidence (HIGH, MEDIUM, LOW).
        details: Additional detection details.
    """

    name: str
    version_hint: str = ""
    known_vulnerabilities: list[KnownVulnerability] = field(default_factory=list)
    recommended_rules: list[str] = field(default_factory=list)
    confidence: str = "HIGH"
    details: dict[str, Any] = field(default_factory=dict)

    def to_dict(self) -> dict[str, Any]:
        return {
            "name": self.name,
            "version_hint": self.version_hint,
            "known_vulnerabilities": [v.to_dict() for v in self.known_vulnerabilities],
            "recommended_rules": self.recommended_rules,
            "confidence": self.confidence,
            "details": self.details,
        }


# ---------------------------------------------------------------------------
# Framework Definitions
# ---------------------------------------------------------------------------

# Known vulnerabilities per framework
_FASTMCP_VULNS: list[KnownVulnerability] = [
    KnownVulnerability(
        id="FASTMCP-001",
        title="Auto-exposure of all decorated functions",
        description=(
            "FastMCP auto-exposes all @tool decorated functions without "
            "explicit access control. This can expose internal functions "
            "that were not intended to be public."
        ),
        severity="HIGH",
        cwe="CWE-284",
        fix="Explicitly list tools to expose. Use tool access control decorators.",
    ),
    KnownVulnerability(
        id="FASTMCP-002",
        title="Default permissive CORS in development mode",
        description=(
            "FastMCP development mode enables permissive CORS by default, "
            "which can be accidentally deployed to production."
        ),
        severity="MEDIUM",
        cwe="CWE-942",
        fix="Set explicit CORS origins. Never use development mode in production.",
    ),
    KnownVulnerability(
        id="FASTMCP-003",
        title="Missing rate limiting by default",
        description=(
            "FastMCP does not include rate limiting by default. "
            "Tools are exposed without any request throttling."
        ),
        severity="MEDIUM",
        cwe="CWE-770",
        fix="Add rate limiting middleware or decorators to all exposed tools.",
    ),
]

_MCP_PYTHON_SDK_VULNS: list[KnownVulnerability] = [
    KnownVulnerability(
        id="MCP-SDK-001",
        title="stdio transport without authentication",
        description=(
            "mcp-python-sdk stdio transport does not include authentication. "
            "Any process that can connect via stdio gets full access."
        ),
        severity="MEDIUM",
        cwe="CWE-306",
        fix="Implement authentication layer on top of stdio transport.",
    ),
    KnownVulnerability(
        id="MCP-SDK-002",
        title="Tool descriptions used as prompt context",
        description=(
            "Tool descriptions in mcp-python-sdk are passed directly to the LLM. "
            "Malicious descriptions can influence model behavior (prompt injection)."
        ),
        severity="HIGH",
        cwe="CWE-74",
        fix="Sanitize tool descriptions. Avoid dynamic content in descriptions.",
    ),
]

_TS_SDK_VULNS: list[KnownVulnerability] = [
    KnownVulnerability(
        id="TS-SDK-001",
        title="No input validation in handler types",
        description=(
            "@modelcontextprotocol/sdk TypeScript handlers accept 'any' types "
            "by default. Without explicit Zod/io-ts validation, type safety "
            "is not enforced at runtime."
        ),
        severity="HIGH",
        cwe="CWE-20",
        fix="Use Zod schemas for all tool handler inputs.",
    ),
    KnownVulnerability(
        id="TS-SDK-002",
        title="Error messages may leak internal details",
        description=(
            "Default error handling in @modelcontextprotocol/sdk passes "
            "error messages directly to clients, potentially leaking "
            "internal paths and stack traces."
        ),
        severity="MEDIUM",
        cwe="CWE-209",
        fix="Implement custom error handler that sanitizes error messages.",
    ),
]

_CUSTOM_VULNS: list[KnownVulnerability] = [
    KnownVulnerability(
        id="CUSTOM-001",
        title="Custom MCP implementation without standard security controls",
        description=(
            "Custom MCP implementations lack the security features of "
            "established SDKs (input validation, error handling, transport "
            "security)."
        ),
        severity="HIGH",
        cwe="CWE-693",
        fix="Consider using an established MCP SDK. Implement all OWASP Agentic controls.",
    ),
]

# Recommended rules per framework
_FRAMEWORK_RULES: dict[str, list[str]] = {
    "FastMCP": [
        "PERM-001", "RATE-001", "MCP-001", "MCP-002", "AUTH-001",
        "GEN-MCP-001", "GEN-MCP-003", "GEN-MCP-005",
    ],
    "mcp-python-sdk": [
        "PATH-001", "CMD-001", "CMD-002", "MCP-001", "MCP-002",
        "MCP-003", "MCP-004", "GEN-MCP-001",
    ],
    "@modelcontextprotocol/sdk": [
        "TS-CMD-001", "TS-SQL-001", "TS-PATH-001", "TS-INPUT-001",
        "TS-AUTH-001", "TS-PROMPT-001",
    ],
    "custom": [
        "PATH-001", "CMD-001", "CMD-002", "AUTH-001", "INPUT-001",
        "RATE-001", "SSRF-001", "DESER-001",
    ],
}


# ---------------------------------------------------------------------------
# Detection Patterns
# ---------------------------------------------------------------------------

_FASTMCP_PATTERNS: list[re.Pattern[str]] = [
    re.compile(r"from\s+fastmcp\s+import", re.IGNORECASE),
    re.compile(r"import\s+fastmcp", re.IGNORECASE),
    re.compile(r"FastMCP\s*\("),
    re.compile(r"@\w+\.tool\b"),
]

_MCP_PYTHON_PATTERNS: list[re.Pattern[str]] = [
    re.compile(r"from\s+mcp\.server\s+import"),
    re.compile(r"from\s+mcp\s+import\s+.*Server"),
    re.compile(r"stdio_server\s*\("),
    re.compile(r"mcp\.server\.Server\s*\("),
]

_TS_SDK_PATTERNS: list[re.Pattern[str]] = [
    re.compile(r"@modelcontextprotocol/sdk"),
    re.compile(r"from\s+['\"]@modelcontextprotocol"),
    re.compile(r"import\s+.*from\s+['\"]@modelcontextprotocol"),
    re.compile(r"McpServer\s*\("),
    re.compile(r"createServer\s*\("),
]

_VERSION_PATTERNS: dict[str, re.Pattern[str]] = {
    "fastmcp": re.compile(r"fastmcp[=<>~!]+([0-9]+\.[0-9]+(?:\.[0-9]+)?)"),
    "mcp": re.compile(r"mcp[=<>~!]+([0-9]+\.[0-9]+(?:\.[0-9]+)?)"),
    "mcp_sdk": re.compile(
        r"@modelcontextprotocol/sdk[\"']?\s*:\s*[\"']?\^?~?([0-9]+\.[0-9]+(?:\.[0-9]+)?)"
    ),
}


# ---------------------------------------------------------------------------
# Fingerprinting Functions
# ---------------------------------------------------------------------------

def fingerprint_server(source_code: str) -> FrameworkInfo:
    """Detect which MCP framework a server uses from its source code.

    Analyzes source code for import patterns, SDK usage, and framework
    markers to determine the MCP framework in use.

    Args:
        source_code: Source code string to analyze.

    Returns:
        FrameworkInfo with detected framework, vulnerabilities, and rules.
    """
    # Score each framework
    scores: dict[str, int] = {
        "FastMCP": 0,
        "mcp-python-sdk": 0,
        "@modelcontextprotocol/sdk": 0,
    }
    matched_patterns: dict[str, list[str]] = {k: [] for k in scores}

    for pat in _FASTMCP_PATTERNS:
        match = pat.search(source_code)
        if match:
            scores["FastMCP"] += 1
            matched_patterns["FastMCP"].append(match.group())

    for pat in _MCP_PYTHON_PATTERNS:
        match = pat.search(source_code)
        if match:
            scores["mcp-python-sdk"] += 1
            matched_patterns["mcp-python-sdk"].append(match.group())

    for pat in _TS_SDK_PATTERNS:
        match = pat.search(source_code)
        if match:
            scores["@modelcontextprotocol/sdk"] += 1
            matched_patterns["@modelcontextprotocol/sdk"].append(match.group())

    # Find winner
    max_score = max(scores.values())
    if max_score == 0:
        # No known framework detected
        return FrameworkInfo(
            name="custom",
            version_hint="",
            known_vulnerabilities=list(_CUSTOM_VULNS),
            recommended_rules=_FRAMEWORK_RULES["custom"],
            confidence="LOW",
            details={"reason": "No known MCP framework patterns detected"},
        )

    winner = max(scores, key=lambda k: scores[k])

    # Determine confidence
    if max_score >= 3:
        confidence = "HIGH"
    elif max_score >= 2:
        confidence = "MEDIUM"
    else:
        confidence = "LOW"

    # Detect version hint
    version_hint = _detect_version(source_code, winner)

    # Get known vulns and rules
    vulns_map = {
        "FastMCP": _FASTMCP_VULNS,
        "mcp-python-sdk": _MCP_PYTHON_SDK_VULNS,
        "@modelcontextprotocol/sdk": _TS_SDK_VULNS,
    }

    return FrameworkInfo(
        name=winner,
        version_hint=version_hint,
        known_vulnerabilities=list(vulns_map.get(winner, [])),
        recommended_rules=_FRAMEWORK_RULES.get(winner, []),
        confidence=confidence,
        details={
            "score": max_score,
            "matched_patterns": matched_patterns[winner],
        },
    )


def fingerprint_file(filepath: Path) -> FrameworkInfo:
    """Fingerprint an MCP server from a file path.

    Args:
        filepath: Path to the source file.

    Returns:
        FrameworkInfo for the detected framework.
    """
    try:
        source = filepath.read_text(encoding="utf-8", errors="replace")
    except (OSError, PermissionError):
        return FrameworkInfo(
            name="unknown",
            confidence="LOW",
            details={"error": f"Could not read file: {filepath}"},
        )
    return fingerprint_server(source)


def fingerprint_directory(dirpath: Path) -> list[tuple[Path, FrameworkInfo]]:
    """Fingerprint all source files in a directory.

    Args:
        dirpath: Directory to scan.

    Returns:
        List of (filepath, FrameworkInfo) tuples for detected servers.
    """
    results: list[tuple[Path, FrameworkInfo]] = []
    extensions = {".py", ".ts", ".js", ".tsx", ".jsx", ".mjs"}
    skip_dirs = {"node_modules", ".git", "__pycache__", ".venv", "venv"}

    if not dirpath.is_dir():
        return results

    for item in sorted(dirpath.rglob("*")):
        if item.is_file() and item.suffix in extensions:
            # Skip excluded dirs
            if any(part in skip_dirs for part in item.parts):
                continue
            info = fingerprint_file(item)
            if info.name != "unknown":
                results.append((item, info))

    return results


def _detect_version(source: str, framework: str) -> str:
    """Try to detect framework version from source or config patterns.

    Args:
        source: Source code to search.
        framework: Framework name to search version for.

    Returns:
        Version string if detected, empty string otherwise.
    """
    key_map = {
        "FastMCP": "fastmcp",
        "mcp-python-sdk": "mcp",
        "@modelcontextprotocol/sdk": "mcp_sdk",
    }
    key = key_map.get(framework)
    if key and key in _VERSION_PATTERNS:
        match = _VERSION_PATTERNS[key].search(source)
        if match:
            return match.group(1)
    return ""


# ---------------------------------------------------------------------------
# CLI Formatting
# ---------------------------------------------------------------------------

def format_fingerprint_text(info: FrameworkInfo, filepath: str = "") -> str:
    """Format fingerprint result as human-readable text.

    Args:
        info: FrameworkInfo to format.
        filepath: Optional filepath for context.

    Returns:
        Formatted text string.
    """
    lines: list[str] = []
    lines.append(f"\n{'=' * 60}")
    lines.append("  mcpaudit -- MCP Server Fingerprint")
    lines.append(f"{'=' * 60}")

    if filepath:
        lines.append(f"  File:       {filepath}")
    lines.append(f"  Framework:  {info.name}")
    if info.version_hint:
        lines.append(f"  Version:    {info.version_hint}")
    lines.append(f"  Confidence: {info.confidence}")

    if info.known_vulnerabilities:
        lines.append(f"\n  Known Vulnerabilities ({len(info.known_vulnerabilities)}):")
        for vuln in info.known_vulnerabilities:
            lines.append(f"    [{vuln.severity}] {vuln.id}: {vuln.title}")
            if vuln.fix:
                lines.append(f"           Fix: {vuln.fix}")

    if info.recommended_rules:
        lines.append(f"\n  Recommended Rules: {', '.join(info.recommended_rules)}")

    lines.append(f"\n{'=' * 60}\n")
    return "\n".join(lines)


def format_fingerprint_json(info: FrameworkInfo, filepath: str = "") -> str:
    """Format fingerprint result as JSON.

    Args:
        info: FrameworkInfo to format.
        filepath: Optional filepath for context.

    Returns:
        JSON string.
    """
    import json
    data = info.to_dict()
    if filepath:
        data["file"] = filepath
    return json.dumps(data, indent=2, ensure_ascii=False)


# ---------------------------------------------------------------------------
# MCPFingerprint — Extended fingerprint with transport, tools, resources
# ---------------------------------------------------------------------------

@dataclass
class MCPFingerprint:
    """Extended fingerprint for an MCP server file or directory.

    Combines framework detection with transport, tool/resource enumeration,
    and version hints for a comprehensive server profile.
    """

    framework: str = "unknown"
    language: str = "unknown"
    version_hints: list[str] = field(default_factory=list)
    transport: str = "unknown"
    capabilities: list[str] = field(default_factory=list)
    tool_count: int = 0
    resource_count: int = 0
    tools: list[str] = field(default_factory=list)
    resources: list[str] = field(default_factory=list)
    file_path: str = ""
    errors: list[str] = field(default_factory=list)

    def to_dict(self) -> dict[str, Any]:
        """Serialize to dictionary."""
        return {
            "framework": self.framework,
            "language": self.language,
            "version_hints": self.version_hints,
            "transport": self.transport,
            "capabilities": self.capabilities,
            "tool_count": self.tool_count,
            "resource_count": self.resource_count,
            "tools": self.tools,
            "resources": self.resources,
            "file_path": self.file_path,
        }


# Transport detection patterns
_STDIO_PATTERN = re.compile(
    r"stdio_server|StdioServerTransport|"
    r"transport.*stdio|type.*stdio",
    re.IGNORECASE,
)
_SSE_PATTERN = re.compile(
    r"sse_server|SSEServerTransport|SseServerTransport|"
    r"transport.*sse|type.*sse",
    re.IGNORECASE,
)
_HTTP_PATTERN = re.compile(
    r"streamable.http|StreamableHTTP|"
    r"transport.*http|type.*streamable",
    re.IGNORECASE,
)

# Tool / resource extraction patterns
_PY_TOOL_NAME = re.compile(
    r'@(?:mcp\.tool|app\.tool|server\.tool|tool)\s*\(\s*'
    r'(?:name\s*=\s*)?["\']([^"\']+)["\']'
)
_PY_TOOL_DEF = re.compile(
    r'@(?:mcp\.tool|app\.tool|server\.tool|tool)\s*\(.*?\)\s*\n\s*'
    r'(?:async\s+)?def\s+(\w+)',
    re.DOTALL,
)
_TS_TOOL_NAME = re.compile(
    r'\.tool\s*\(\s*["\']([^"\']+)["\']'
)
_PY_RESOURCE_NAME = re.compile(
    r'@(?:mcp\.resource|app\.resource|server\.resource|resource)\s*\(\s*'
    r'["\']([^"\']+)["\']'
)
_TS_RESOURCE_NAME = re.compile(
    r'\.resource\s*\(\s*["\']([^"\']+)["\']'
)

# Capability patterns
_CAPABILITY_PATTERNS: list[tuple[str, re.Pattern[str]]] = [
    ("tools", re.compile(r"\.tool\s*\(|@.*tool", re.IGNORECASE)),
    ("resources", re.compile(r"\.resource\s*\(|@.*resource", re.IGNORECASE)),
    ("prompts", re.compile(r"\.prompt\s*\(|@.*prompt", re.IGNORECASE)),
    ("logging", re.compile(r"logging\.getLogger|logger\s*=", re.IGNORECASE)),
    ("sampling", re.compile(r"sampling|create_message", re.IGNORECASE)),
]

_PY_EXTS = frozenset({".py", ".pyw"})
_TS_EXTS = frozenset({".ts", ".tsx", ".js", ".jsx", ".mjs"})
_SKIP_DIRS = frozenset({
    "node_modules", "__pycache__", ".git", ".venv", "venv",
})


def _detect_lang(filepath: Path) -> str:
    """Detect language from file extension."""
    ext = filepath.suffix.lower()
    if ext in _PY_EXTS:
        return "python"
    if ext in _TS_EXTS:
        return "typescript"
    return "unknown"


def _detect_transport(content: str) -> str:
    """Detect MCP transport type from source content."""
    if _STDIO_PATTERN.search(content):
        return "stdio"
    if _SSE_PATTERN.search(content):
        return "sse"
    if _HTTP_PATTERN.search(content):
        return "streamable-http"
    return "unknown"


def _detect_capabilities(content: str) -> list[str]:
    """Detect MCP capabilities from source code."""
    return [
        name for name, pat in _CAPABILITY_PATTERNS
        if pat.search(content)
    ]


def _extract_tools(content: str, language: str) -> list[str]:
    """Extract tool names from source code."""
    tools: list[str] = []
    if language == "python":
        for m in _PY_TOOL_NAME.finditer(content):
            tools.append(m.group(1))
        for m in _PY_TOOL_DEF.finditer(content):
            name = m.group(1)
            if name not in tools:
                tools.append(name)
    else:
        for m in _TS_TOOL_NAME.finditer(content):
            tools.append(m.group(1))
    return tools


def _extract_resources(content: str, language: str) -> list[str]:
    """Extract resource URIs from source code."""
    if language == "python":
        return [m.group(1) for m in _PY_RESOURCE_NAME.finditer(content)]
    return [m.group(1) for m in _TS_RESOURCE_NAME.finditer(content)]


def _extract_versions(content: str, framework: str) -> list[str]:
    """Extract version hints for the detected framework."""
    hints: list[str] = []
    for key, pat in _VERSION_PATTERNS.items():
        for m in pat.finditer(content):
            hints.append(m.group(1))
    return hints


def _fw_name_to_mcpfp(fw_name: str) -> str:
    """Normalize FrameworkInfo name to MCPFingerprint framework field."""
    mapping = {
        "FastMCP": "fastmcp",
        "mcp-python-sdk": "mcp-sdk",
        "@modelcontextprotocol/sdk": "mcp-sdk",
        "custom": "custom",
        "unknown": "unknown",
    }
    return mapping.get(fw_name, fw_name.lower())


def fingerprint_path(path: str | Path) -> MCPFingerprint:
    """Fingerprint an MCP server from a file or directory path.

    Args:
        path: Path to a source file or directory.

    Returns:
        MCPFingerprint with detected characteristics.
    """
    target = Path(path)
    fp = MCPFingerprint(file_path=str(target))

    if not target.exists():
        fp.errors.append(f"Path does not exist: {target}")
        return fp

    if target.is_dir():
        return _fingerprint_dir(target)

    return _fingerprint_single(target)


def _fingerprint_single(filepath: Path) -> MCPFingerprint:
    """Fingerprint a single source file."""
    fp = MCPFingerprint(file_path=str(filepath))

    try:
        content = filepath.read_text(encoding="utf-8", errors="replace")
    except (OSError, PermissionError) as exc:
        fp.errors.append(f"Cannot read file: {exc}")
        return fp

    fp.language = _detect_lang(filepath)

    # Use existing fingerprint_server for framework detection
    fw_info = fingerprint_server(content)
    fp.framework = _fw_name_to_mcpfp(fw_info.name)

    fp.transport = _detect_transport(content)
    fp.capabilities = _detect_capabilities(content)
    fp.version_hints = _extract_versions(content, fw_info.name)
    fp.tools = _extract_tools(content, fp.language)
    fp.resources = _extract_resources(content, fp.language)
    fp.tool_count = len(fp.tools)
    fp.resource_count = len(fp.resources)

    return fp


def _fingerprint_dir(directory: Path) -> MCPFingerprint:
    """Fingerprint a directory by aggregating all source files."""
    combined = MCPFingerprint(file_path=str(directory))
    all_tools: list[str] = []
    all_resources: list[str] = []
    all_caps: list[str] = []
    all_versions: list[str] = []

    for filepath in sorted(directory.rglob("*")):
        if not filepath.is_file():
            continue
        if filepath.suffix.lower() not in (_PY_EXTS | _TS_EXTS):
            continue
        if any(p in filepath.parts for p in _SKIP_DIRS):
            continue

        fp = _fingerprint_single(filepath)
        if fp.framework not in ("unknown", "custom") and combined.framework in ("unknown", "custom"):
            combined.framework = fp.framework
            combined.language = fp.language
        if fp.transport != "unknown" and combined.transport == "unknown":
            combined.transport = fp.transport
        all_tools.extend(fp.tools)
        all_resources.extend(fp.resources)
        all_caps.extend(fp.capabilities)
        all_versions.extend(fp.version_hints)

    combined.tools = list(dict.fromkeys(all_tools))
    combined.resources = list(dict.fromkeys(all_resources))
    combined.capabilities = list(dict.fromkeys(all_caps))
    combined.version_hints = list(dict.fromkeys(all_versions))
    combined.tool_count = len(combined.tools)
    combined.resource_count = len(combined.resources)

    if combined.framework == "unknown":
        combined.errors.append("No MCP server patterns detected")

    return combined


def format_mcpfingerprint_text(fp: MCPFingerprint) -> str:
    """Format MCPFingerprint as human-readable text."""
    lines: list[str] = []
    lines.append(f"\n{'=' * 50}")
    lines.append("  mcpaudit -- MCP Server Fingerprint")
    lines.append(f"{'=' * 50}")
    lines.append(f"  File:       {fp.file_path}")
    lines.append(f"  Framework:  {fp.framework}")
    lines.append(f"  Language:   {fp.language}")
    lines.append(f"  Transport:  {fp.transport}")
    lines.append(f"  Tools:      {fp.tool_count}")
    lines.append(f"  Resources:  {fp.resource_count}")

    if fp.version_hints:
        lines.append(f"  Versions:   {', '.join(fp.version_hints)}")
    if fp.capabilities:
        lines.append(f"  Capabilities: {', '.join(fp.capabilities)}")
    if fp.tools:
        lines.append("  Tool names:")
        for t in fp.tools:
            lines.append(f"    - {t}")
    if fp.resources:
        lines.append("  Resource URIs:")
        for r in fp.resources:
            lines.append(f"    - {r}")
    if fp.errors:
        lines.append("  Errors:")
        for e in fp.errors:
            lines.append(f"    ! {e}")

    lines.append(f"{'=' * 50}\n")
    return "\n".join(lines)


def format_mcpfingerprint_json(fp: MCPFingerprint) -> str:
    """Format MCPFingerprint as JSON."""
    import json as _json
    return _json.dumps(fp.to_dict(), indent=2, ensure_ascii=False)



"""Tests for MCP Vulnerability Database v2."""

from __future__ import annotations

import json
import tempfile
from pathlib import Path

import pytest

from mcpaudit.vuln_db_v2 import (
    EnrichedFinding,
    VulnDatabase,
    VulnEntry,
    VulnSearchResult,
    VulnSeverity,
    _build_vuln_db,
    _calculate_risk_score,
    export_db_json,
    export_db_to_file,
    format_enriched_json,
    format_enriched_text,
    format_vuln_db_json,
    format_vuln_db_text,
    import_db_from_file,
    import_db_json,
    match_findings,
)


# --- VulnSeverity ---

class TestVulnSeverity:
    def test_all_values(self):
        assert VulnSeverity.CRITICAL.value == "CRITICAL"
        assert VulnSeverity.HIGH.value == "HIGH"
        assert VulnSeverity.MEDIUM.value == "MEDIUM"
        assert VulnSeverity.LOW.value == "LOW"
        assert VulnSeverity.INFO.value == "INFO"

    def test_from_string_valid(self):
        assert VulnSeverity.from_string("critical") == VulnSeverity.CRITICAL
        assert VulnSeverity.from_string("HIGH") == VulnSeverity.HIGH
        assert VulnSeverity.from_string("Medium") == VulnSeverity.MEDIUM

    def test_from_string_invalid(self):
        with pytest.raises(ValueError, match="Invalid severity"):
            VulnSeverity.from_string("unknown")

    def test_enum_iteration(self):
        assert len(list(VulnSeverity)) == 5


# --- VulnEntry ---

class TestVulnEntry:
    def test_creation(self):
        e = VulnEntry(id="V-001", title="Test", description="Desc",
                      severity=VulnSeverity.HIGH)
        assert e.id == "V-001"
        assert e.severity == VulnSeverity.HIGH

    def test_to_dict(self):
        e = VulnEntry(id="V-001", title="T", description="D",
                      severity=VulnSeverity.MEDIUM, tags=["a"], rule_ids=["R-001"],
                      cvss_score=5.0)
        d = e.to_dict()
        assert d["id"] == "V-001"
        assert d["severity"] == "MEDIUM"
        assert d["cvss_score"] == 5.0

    def test_from_dict(self):
        data = {"id": "V-002", "title": "T2", "description": "D2",
                "severity": "HIGH", "tags": ["x"], "cvss_score": 7.5}
        e = VulnEntry.from_dict(data)
        assert e.id == "V-002"
        assert e.severity == VulnSeverity.HIGH
        assert e.cvss_score == 7.5

    def test_from_dict_defaults(self):
        data = {"id": "V-003", "title": "T3", "description": "D3", "severity": "LOW"}
        e = VulnEntry.from_dict(data)
        assert e.tags == []
        assert e.cwe_ids == []
        assert e.exploitability == "UNKNOWN"
        assert e.cvss_score == 0.0

    def test_roundtrip(self):
        e = VulnEntry(id="V-R", title="RT", description="Roundtrip",
                      severity=VulnSeverity.CRITICAL, cwe_ids=["CWE-78"],
                      tags=["test"], published="2024-01-01", cvss_score=9.5)
        d = e.to_dict()
        e2 = VulnEntry.from_dict(d)
        assert e2.id == e.id
        assert e2.severity == e.severity
        assert e2.cvss_score == e.cvss_score

    def test_affected_patterns(self):
        e = VulnEntry(id="V-P", title="T", description="D",
                      severity=VulnSeverity.HIGH,
                      affected_patterns=["subprocess.call(", "os.system("])
        assert len(e.affected_patterns) == 2

    def test_affected_languages(self):
        e = VulnEntry(id="V-L", title="T", description="D",
                      severity=VulnSeverity.HIGH,
                      affected_languages=["python", "go"])
        assert "python" in e.affected_languages


# --- EnrichedFinding ---

class TestEnrichedFinding:
    def test_not_enriched(self):
        ef = EnrichedFinding(finding={"rule_id": "X"})
        assert not ef.is_enriched
        assert ef.max_severity == "INFO"

    def test_enriched_with_vulns(self):
        v = VulnEntry(id="V-1", title="T", description="D",
                      severity=VulnSeverity.CRITICAL)
        ef = EnrichedFinding(finding={"rule_id": "X"}, matched_vulns=[v])
        assert ef.is_enriched
        assert ef.max_severity == "CRITICAL"

    def test_max_severity_multiple(self):
        v1 = VulnEntry(id="V-1", title="T", description="D", severity=VulnSeverity.LOW)
        v2 = VulnEntry(id="V-2", title="T", description="D", severity=VulnSeverity.HIGH)
        ef = EnrichedFinding(finding={}, matched_vulns=[v1, v2])
        assert ef.max_severity == "HIGH"

    def test_to_dict(self):
        ef = EnrichedFinding(finding={"rule_id": "R"}, risk_score=42.0)
        d = ef.to_dict()
        assert d["risk_score"] == 42.0
        assert d["is_enriched"] is False

    def test_enrichment_notes(self):
        ef = EnrichedFinding(finding={}, enrichment_notes=["note1", "note2"])
        assert len(ef.enrichment_notes) == 2

    def test_finding_severity_fallback(self):
        ef = EnrichedFinding(finding={"severity": "HIGH"})
        assert ef.max_severity == "HIGH"


# --- VulnSearchResult ---

class TestVulnSearchResult:
    def test_empty(self):
        sr = VulnSearchResult()
        assert sr.count == 0

    def test_with_entries(self):
        e = VulnEntry(id="V", title="T", description="D", severity=VulnSeverity.LOW)
        sr = VulnSearchResult(entries=[e], query="test", total_in_db=10)
        assert sr.count == 1
        assert sr.total_in_db == 10

    def test_to_dict(self):
        sr = VulnSearchResult(query="q", total_in_db=5)
        d = sr.to_dict()
        assert d["query"] == "q"
        assert d["results_count"] == 0


# --- _build_vuln_db ---

class TestBuildVulnDB:
    def test_count(self):
        db = _build_vuln_db()
        assert len(db) >= 30

    def test_all_have_ids(self):
        for e in _build_vuln_db():
            assert e.id.startswith("MCP-VULN-")

    def test_all_have_severity(self):
        for e in _build_vuln_db():
            assert isinstance(e.severity, VulnSeverity)

    def test_all_have_cwe(self):
        for e in _build_vuln_db():
            assert len(e.cwe_ids) > 0

    def test_all_have_published(self):
        for e in _build_vuln_db():
            assert e.published != ""

    def test_unique_ids(self):
        ids = [e.id for e in _build_vuln_db()]
        assert len(ids) == len(set(ids))


# --- VulnDatabase ---

class TestVulnDatabase:
    def test_default_init(self):
        db = VulnDatabase()
        assert db.count >= 30

    def test_custom_entries(self):
        e = VulnEntry(id="V-1", title="T", description="D", severity=VulnSeverity.LOW)
        db = VulnDatabase(entries=[e])
        assert db.count == 1

    def test_get_existing(self):
        db = VulnDatabase()
        e = db.get("MCP-VULN-001")
        assert e is not None
        assert e.id == "MCP-VULN-001"

    def test_get_missing(self):
        db = VulnDatabase()
        assert db.get("NONEXISTENT") is None

    def test_list_all(self):
        db = VulnDatabase()
        all_entries = db.list_all()
        assert len(all_entries) == db.count

    def test_search_keyword(self):
        db = VulnDatabase()
        results = db.search("injection")
        assert len(results) > 0

    def test_search_cwe(self):
        db = VulnDatabase()
        results = db.search("CWE-78")
        assert len(results) > 0

    def test_search_no_results(self):
        db = VulnDatabase()
        results = db.search("zzzznonexistentzzzz")
        assert len(results) == 0

    def test_search_case_insensitive(self):
        db = VulnDatabase()
        r1 = db.search("injection")
        r2 = db.search("INJECTION")
        assert len(r1) == len(r2)

    def test_filter_by_severity_enum(self):
        db = VulnDatabase()
        critical = db.filter_by_severity(VulnSeverity.CRITICAL)
        assert len(critical) > 0
        assert all(e.severity == VulnSeverity.CRITICAL for e in critical)

    def test_filter_by_severity_string(self):
        db = VulnDatabase()
        high = db.filter_by_severity("high")
        assert len(high) > 0

    def test_filter_by_cwe(self):
        db = VulnDatabase()
        results = db.filter_by_cwe("CWE-78")
        assert len(results) > 0
        assert all("CWE-78" in e.cwe_ids for e in results)

    def test_filter_by_cwe_no_prefix(self):
        db = VulnDatabase()
        results = db.filter_by_cwe("78")
        assert len(results) > 0

    def test_filter_by_tag(self):
        db = VulnDatabase()
        results = db.filter_by_tag("injection")
        assert len(results) > 0

    def test_filter_by_tag_case_insensitive(self):
        db = VulnDatabase()
        r1 = db.filter_by_tag("owasp")
        r2 = db.filter_by_tag("OWASP")
        assert len(r1) == len(r2)

    def test_filter_by_date_range(self):
        db = VulnDatabase()
        results = db.filter_by_date_range(date_from="2024-07-01", date_to="2024-09-30")
        assert len(results) > 0
        for e in results:
            assert "2024-07" <= e.published <= "2024-09-30"

    def test_filter_by_date_from_only(self):
        db = VulnDatabase()
        results = db.filter_by_date_range(date_from="2025-01-01")
        assert len(results) > 0

    def test_filter_by_date_to_only(self):
        db = VulnDatabase()
        results = db.filter_by_date_range(date_to="2024-06-30")
        assert len(results) > 0

    def test_vulns_for_rule(self):
        db = VulnDatabase()
        results = db.vulns_for_rule("CMD-001")
        assert len(results) > 0

    def test_vulns_for_rule_missing(self):
        db = VulnDatabase()
        results = db.vulns_for_rule("NONEXISTENT-999")
        assert results == []

    def test_add_entry(self):
        db = VulnDatabase(entries=[])
        e = VulnEntry(id="NEW-001", title="New", description="D",
                      severity=VulnSeverity.LOW, rule_ids=["R-NEW"])
        db.add_entry(e)
        assert db.count == 1
        assert db.get("NEW-001") is not None
        assert len(db.vulns_for_rule("R-NEW")) == 1

    def test_remove_entry(self):
        e = VulnEntry(id="DEL-001", title="Del", description="D",
                      severity=VulnSeverity.LOW, rule_ids=["R-DEL"])
        db = VulnDatabase(entries=[e])
        assert db.remove_entry("DEL-001") is True
        assert db.count == 0
        assert db.get("DEL-001") is None

    def test_remove_nonexistent(self):
        db = VulnDatabase(entries=[])
        assert db.remove_entry("NOPE") is False

    def test_search_advanced_query(self):
        db = VulnDatabase()
        r = db.search_advanced(query="injection")
        assert r.count > 0
        assert r.total_in_db == db.count

    def test_search_advanced_severity(self):
        db = VulnDatabase()
        r = db.search_advanced(severity="CRITICAL")
        assert r.count > 0

    def test_search_advanced_cwe(self):
        db = VulnDatabase()
        r = db.search_advanced(cwe="78")
        assert r.count > 0

    def test_search_advanced_language(self):
        db = VulnDatabase()
        r = db.search_advanced(language="python")
        assert r.count > 0

    def test_search_advanced_tag(self):
        db = VulnDatabase()
        r = db.search_advanced(tag="owasp")
        assert r.count > 0

    def test_search_advanced_date_range(self):
        db = VulnDatabase()
        r = db.search_advanced(date_from="2024-07-01", date_to="2024-12-31")
        assert r.count > 0

    def test_search_advanced_combined(self):
        db = VulnDatabase()
        r = db.search_advanced(severity="HIGH", language="python")
        assert r.count > 0

    def test_search_advanced_no_results(self):
        db = VulnDatabase()
        r = db.search_advanced(query="zzznonexistent", severity="INFO")
        assert r.count == 0


# --- match_findings ---

class TestMatchFindings:
    def test_match_by_rule_id(self):
        findings = [{"rule_id": "CMD-001", "title": "Command injection", "file": "a.py", "line": 1}]
        enriched = match_findings(findings)
        assert len(enriched) == 1
        assert enriched[0].is_enriched

    def test_no_match(self):
        findings = [{"rule_id": "NONEXISTENT-999", "title": "Nothing", "file": "b.py", "line": 1}]
        enriched = match_findings(findings)
        assert len(enriched) == 1
        assert not enriched[0].is_enriched

    def test_match_with_custom_db(self):
        e = VulnEntry(id="CUSTOM-1", title="Custom Vuln", description="D",
                      severity=VulnSeverity.HIGH, rule_ids=["CUSTOM-R"])
        db = VulnDatabase(entries=[e])
        findings = [{"rule_id": "CUSTOM-R", "title": "T", "file": "c.py", "line": 5}]
        enriched = match_findings(findings, db)
        assert enriched[0].is_enriched
        assert enriched[0].matched_vulns[0].id == "CUSTOM-1"

    def test_enrichment_notes(self):
        findings = [{"rule_id": "CMD-001", "title": "Cmd inj", "file": "a.py", "line": 1}]
        enriched = match_findings(findings)
        if enriched[0].is_enriched:
            assert len(enriched[0].enrichment_notes) > 0

    def test_risk_score_calculated(self):
        findings = [{"rule_id": "CMD-001", "title": "T", "file": "a.py", "line": 1}]
        enriched = match_findings(findings)
        assert enriched[0].risk_score > 0

    def test_multiple_findings(self):
        findings = [
            {"rule_id": "CMD-001", "title": "T1", "file": "a.py", "line": 1},
            {"rule_id": "NET-001", "title": "T2", "file": "b.py", "line": 2},
        ]
        enriched = match_findings(findings)
        assert len(enriched) == 2

    def test_empty_findings(self):
        enriched = match_findings([])
        assert enriched == []


# --- _calculate_risk_score ---

class TestCalculateRiskScore:
    def test_no_vulns_critical(self):
        s = _calculate_risk_score({"severity": "CRITICAL"}, [])
        assert s == 40.0

    def test_no_vulns_info(self):
        s = _calculate_risk_score({"severity": "INFO"}, [])
        assert s == 5.0

    def test_with_high_cvss(self):
        v = VulnEntry(id="V", title="T", description="D",
                      severity=VulnSeverity.CRITICAL, cvss_score=10.0,
                      exploitability="HIGH")
        s = _calculate_risk_score({}, [v])
        assert s >= 70.0

    def test_capped_at_100(self):
        vulns = [VulnEntry(id=f"V-{i}", title="T", description="D",
                           severity=VulnSeverity.CRITICAL, cvss_score=10.0,
                           exploitability="HIGH") for i in range(20)]
        s = _calculate_risk_score({}, vulns)
        assert s <= 100.0


# --- Import/Export ---

class TestImportExport:
    def test_export_json(self):
        db = VulnDatabase()
        j = export_db_json(db)
        data = json.loads(j)
        assert data["version"] == "2.0"
        assert data["count"] == db.count

    def test_import_json(self):
        db = VulnDatabase()
        j = export_db_json(db)
        db2 = import_db_json(j)
        assert db2.count == db.count

    def test_roundtrip_json(self):
        e = VulnEntry(id="RT-1", title="Roundtrip", description="D",
                      severity=VulnSeverity.HIGH, cvss_score=7.0)
        db = VulnDatabase(entries=[e])
        j = export_db_json(db)
        db2 = import_db_json(j)
        assert db2.get("RT-1") is not None
        assert db2.get("RT-1").cvss_score == 7.0

    def test_export_to_file(self):
        db = VulnDatabase(entries=[])
        with tempfile.NamedTemporaryFile(suffix=".json", delete=False) as f:
            export_db_to_file(db, f.name)
            content = Path(f.name).read_text()
            assert "entries" in content

    def test_import_from_file(self):
        e = VulnEntry(id="FILE-1", title="File", description="D",
                      severity=VulnSeverity.LOW)
        db = VulnDatabase(entries=[e])
        with tempfile.NamedTemporaryFile(suffix=".json", delete=False, mode="w") as f:
            f.write(export_db_json(db))
            f.flush()
            db2 = import_db_from_file(f.name)
            assert db2.count == 1


# --- Formatting ---

class TestFormatting:
    def test_format_text(self):
        db = VulnDatabase()
        text = format_vuln_db_text(db)
        assert "Vulnerability Database" in text
        assert "MCP-VULN-001" in text

    def test_format_json(self):
        db = VulnDatabase()
        j = format_vuln_db_json(db)
        data = json.loads(j)
        assert "entries" in data

    def test_format_enriched_text(self):
        ef = EnrichedFinding(finding={"rule_id": "R", "title": "T"}, risk_score=50.0)
        text = format_enriched_text([ef])
        assert "Enriched" in text
        assert "50.0" in text

    def test_format_enriched_json(self):
        ef = EnrichedFinding(finding={"rule_id": "R"})
        j = format_enriched_json([ef])
        data = json.loads(j)
        assert len(data) == 1

    def test_format_enriched_empty(self):
        text = format_enriched_text([])
        assert "0" in text



"""Tests for Enhanced Auto-Fix Engine -- v1.3.0 expanded rule coverage.

Tests suggest_fix() on 18 rules (was 5), fix statistics, and CLI
integration for --fix-stats and new fixable rules.

NOTE: This test file contains intentionally vulnerable code STRINGS
for testing the scanner and fixer. These strings reference unsafe
functions (pickle, yaml.load, etc.) as detection signatures -- they
are NOT executed. This is a SECURITY SCANNER test suite.
"""

from __future__ import annotations

import ast
import tempfile
from pathlib import Path

import pytest

from mcpaudit.fixer import (
    FixResult,
    FixStats,
    FixSuggestion,
    apply_fixes,
    build_rule_map,
    compute_fix_stats,
    format_fix_stats,
    get_fixable_rule_ids,
    suggest_fixes,
)
from mcpaudit.models import Finding, Severity
from mcpaudit.rules import get_all_rules


# ---------------------------------------------------------------------------
# Helpers
# ---------------------------------------------------------------------------

def _check(rule, source: str) -> list[Finding]:
    """Run a single rule against source code and return findings."""
    tree = ast.parse(source)
    return rule.check(tree, source)


def _write_temp(content: str, suffix: str = ".py") -> Path:
    """Write content to a temp file and return its path."""
    f = tempfile.NamedTemporaryFile(
        suffix=suffix, mode="w", delete=False, encoding="utf-8",
    )
    f.write(content)
    f.flush()
    f.close()
    return Path(f.name)


# ===========================================================================
# get_fixable_rule_ids
# ===========================================================================

class TestGetFixableRuleIds:
    """Tests for the get_fixable_rule_ids() utility."""

    def test_returns_list(self):
        ids = get_fixable_rule_ids()
        assert isinstance(ids, list)

    def test_at_least_18_fixable(self):
        ids = get_fixable_rule_ids()
        assert len(ids) >= 18

    def test_sorted(self):
        ids = get_fixable_rule_ids()
        assert ids == sorted(ids)

    def test_contains_core_rules(self):
        ids = get_fixable_rule_ids()
        for rule_id in [
            "AUTH-001", "CMD-001", "CMD-002", "CRYPTO-001",
            "DESER-001", "ERR-001", "FILE-001", "INFO-001",
            "LOG-001", "MCP-005", "MCP-006", "PATH-001",
            "PERM-001", "RES-001", "SEC-001", "SQL-001",
            "SSRF-001", "VAL-001",
        ]:
            assert rule_id in ids, f"{rule_id} not found in fixable rules"


# ===========================================================================
# CMD-001: shell=True -> shell=False
# ===========================================================================

class TestFixCMD001:
    """Tests for CMD-001 suggest_fix (shell=True)."""

    def test_basic_shell_true_fix(self):
        from mcpaudit.rules.command_injection import CommandInjectionShellRule
        source = "subprocess.run('ls', shell=True)\n"
        rule = CommandInjectionShellRule()
        findings = _check(rule, source)
        assert len(findings) >= 1
        fix = rule.suggest_fix(source, findings[0])
        assert fix is not None
        assert "shell=False" in fix
        assert "shell=True" not in fix

    def test_shell_true_with_other_args(self):
        from mcpaudit.rules.command_injection import CommandInjectionShellRule
        source = "subprocess.run(cmd, shell=True, timeout=30)\n"
        rule = CommandInjectionShellRule()
        findings = _check(rule, source)
        assert len(findings) >= 1
        fix = rule.suggest_fix(source, findings[0])
        assert "shell=False" in fix
        assert "timeout=30" in fix

    def test_invalid_line_returns_none(self):
        from mcpaudit.rules.command_injection import CommandInjectionShellRule
        rule = CommandInjectionShellRule()
        finding = Finding(
            rule_id="CMD-001", severity=Severity.CRITICAL,
            title="test", description="test",
            file="test.py", line=999,
        )
        fix = rule.suggest_fix("x = 1\n", finding)
        assert fix is None

    def test_no_shell_true_returns_none(self):
        from mcpaudit.rules.command_injection import CommandInjectionShellRule
        rule = CommandInjectionShellRule()
        finding = Finding(
            rule_id="CMD-001", severity=Severity.CRITICAL,
            title="test", description="test",
            file="test.py", line=1,
        )
        fix = rule.suggest_fix("x = 1\n", finding)
        assert fix is None

    def test_integrated_fix(self):
        from mcpaudit.rules.command_injection import CommandInjectionShellRule
        source = "subprocess.run('ls', shell=True)\n"
        f = _write_temp(source)
        rule = CommandInjectionShellRule()
        findings = _check(rule, source)
        for fi in findings:
            fi.file = str(f)
        rm = build_rule_map([rule])
        result = suggest_fixes(str(f), findings, rm)
        assert result.fixable_count >= 1


# ===========================================================================
# CMD-002: dynamic args
# ===========================================================================

class TestFixCMD002:
    """Tests for CMD-002 suggest_fix (dynamic args)."""

    def test_dynamic_fstring_fix(self):
        from mcpaudit.rules.command_injection import CommandInjectionDynamicRule
        source = 'subprocess.run(f"ls {path}")\n'
        rule = CommandInjectionDynamicRule()
        findings = _check(rule, source)
        assert len(findings) >= 1
        fix = rule.suggest_fix(source, findings[0])
        assert fix is not None
        assert "TODO" in fix

    def test_invalid_line_returns_none(self):
        from mcpaudit.rules.command_injection import CommandInjectionDynamicRule
        rule = CommandInjectionDynamicRule()
        finding = Finding(
            rule_id="CMD-002", severity=Severity.HIGH,
            title="test", description="test",
            file="test.py", line=999,
        )
        fix = rule.suggest_fix("x = 1\n", finding)
        assert fix is None


# ===========================================================================
# SQL-001: parameterized query
# ===========================================================================

class TestFixSQL001:
    """Tests for SQL-001 suggest_fix (SQL injection)."""

    def test_basic_sql_injection_fix(self):
        from mcpaudit.rules.sql_injection import SqlInjectionRule
        source = 'cursor.execute(f"SELECT * FROM users WHERE id={uid}")\n'
        rule = SqlInjectionRule()
        findings = _check(rule, source)
        assert len(findings) >= 1
        fix = rule.suggest_fix(source, findings[0])
        assert fix is not None
        assert "parameterized" in fix.lower() or "TODO" in fix

    def test_preserves_indentation(self):
        from mcpaudit.rules.sql_injection import SqlInjectionRule
        source = 'def query():\n    cursor.execute(f"DELETE FROM t WHERE id={x}")\n'
        rule = SqlInjectionRule()
        findings = _check(rule, source)
        assert len(findings) >= 1
        fix = rule.suggest_fix(source, findings[0])
        assert fix is not None
        assert fix.startswith("    ")

    def test_invalid_line_returns_none(self):
        from mcpaudit.rules.sql_injection import SqlInjectionRule
        rule = SqlInjectionRule()
        finding = Finding(
            rule_id="SQL-001", severity=Severity.CRITICAL,
            title="test", description="test",
            file="test.py", line=999,
        )
        fix = rule.suggest_fix("x = 1\n", finding)
        assert fix is None


# ===========================================================================
# DESER-001: unsafe deserialization
# Detection signatures for security scanner -- not actual usage
# ===========================================================================

class TestFixDESER001:
    """Tests for DESER-001 suggest_fix (unsafe deserialization)."""

    def test_yaml_load_to_safe_load(self):
        from mcpaudit.rules.unsafe_deser import UnsafeDeserializationRule
        source = "data = yaml.load(f)\n"
        rule = UnsafeDeserializationRule()
        findings = _check(rule, source)
        assert len(findings) >= 1
        fix = rule.suggest_fix(source, findings[0])
        assert fix is not None
        assert "safe_load" in fix

    def test_invalid_line_returns_none(self):
        from mcpaudit.rules.unsafe_deser import UnsafeDeserializationRule
        rule = UnsafeDeserializationRule()
        finding = Finding(
            rule_id="DESER-001", severity=Severity.CRITICAL,
            title="test", description="test",
            file="test.py", line=999,
        )
        fix = rule.suggest_fix("x = 1\n", finding)
        assert fix is None


# ===========================================================================
# ERR-001: unhandled tool errors
# ===========================================================================

class TestFixERR001:
    """Tests for ERR-001 suggest_fix (error handling)."""

    def test_tool_handler_fix(self):
        from mcpaudit.rules.error_handling import UnhandledToolErrorRule
        source = "async def call_tool(name, args):\n    return process(args)\n"
        rule = UnhandledToolErrorRule()
        findings = _check(rule, source)
        assert len(findings) >= 1
        fix = rule.suggest_fix(source, findings[0])
        assert fix is not None
        assert "try/except" in fix or "TODO" in fix

    def test_preserves_indentation(self):
        from mcpaudit.rules.error_handling import UnhandledToolErrorRule
        source = "class S:\n    async def handle_tool(self, name, args):\n        return process(args)\n"
        rule = UnhandledToolErrorRule()
        findings = _check(rule, source)
        if findings:
            fix = rule.suggest_fix(source, findings[0])
            if fix is not None:
                assert fix.startswith("    ")

    def test_invalid_line_returns_none(self):
        from mcpaudit.rules.error_handling import UnhandledToolErrorRule
        rule = UnhandledToolErrorRule()
        finding = Finding(
            rule_id="ERR-001", severity=Severity.MEDIUM,
            title="test", description="test",
            file="test.py", line=999,
        )
        fix = rule.suggest_fix("x = 1\n", finding)
        assert fix is None


# ===========================================================================
# SSRF-001: URL allowlist
# ===========================================================================

class TestFixSSRF001:
    """Tests for SSRF-001 suggest_fix (SSRF)."""

    def test_requests_get_fix(self):
        from mcpaudit.rules.ssrf import SSRFRule
        source = "response = requests.get(user_url)\n"
        rule = SSRFRule()
        findings = _check(rule, source)
        assert len(findings) >= 1
        fix = rule.suggest_fix(source, findings[0])
        assert fix is not None
        assert "allowlist" in fix.lower() or "TODO" in fix

    def test_invalid_line_returns_none(self):
        from mcpaudit.rules.ssrf import SSRFRule
        rule = SSRFRule()
        finding = Finding(
            rule_id="SSRF-001", severity=Severity.HIGH,
            title="test", description="test",
            file="test.py", line=999,
        )
        fix = rule.suggest_fix("x = 1\n", finding)
        assert fix is None


# ===========================================================================
# LOG-001: sensitive data in logs
# ===========================================================================

class TestFixLOG001:
    """Tests for LOG-001 suggest_fix (sensitive logging)."""

    def test_logging_sensitive_fix(self):
        from mcpaudit.rules.sensitive_logging import SensitiveLoggingRule
        source = "logger.info(f'User: {password}')\n"
        rule = SensitiveLoggingRule()
        findings = _check(rule, source)
        assert len(findings) >= 1
        fix = rule.suggest_fix(source, findings[0])
        assert fix is not None
        assert "Redact" in fix or "TODO" in fix

    def test_invalid_line_returns_none(self):
        from mcpaudit.rules.sensitive_logging import SensitiveLoggingRule
        rule = SensitiveLoggingRule()
        finding = Finding(
            rule_id="LOG-001", severity=Severity.MEDIUM,
            title="test", description="test",
            file="test.py", line=999,
        )
        fix = rule.suggest_fix("x = 1\n", finding)
        assert fix is None


# ===========================================================================
# FILE-001: unsafe file write
# ===========================================================================

class TestFixFILE001:
    """Tests for FILE-001 suggest_fix (unsafe file write)."""

    def test_open_write_fix(self):
        from mcpaudit.rules.unsafe_file_write import UnsafeFileWriteRule
        source = "f = open(user_path, 'w')\n"
        rule = UnsafeFileWriteRule()
        findings = _check(rule, source)
        assert len(findings) >= 1
        fix = rule.suggest_fix(source, findings[0])
        assert fix is not None
        assert "Validate" in fix or "realpath" in fix or "TODO" in fix

    def test_invalid_line_returns_none(self):
        from mcpaudit.rules.unsafe_file_write import UnsafeFileWriteRule
        rule = UnsafeFileWriteRule()
        finding = Finding(
            rule_id="FILE-001", severity=Severity.HIGH,
            title="test", description="test",
            file="test.py", line=999,
        )
        fix = rule.suggest_fix("x = 1\n", finding)
        assert fix is None


# ===========================================================================
# RES-001: resource exhaustion (timeout)
# ===========================================================================

class TestFixRES001:
    """Tests for RES-001 suggest_fix (resource exhaustion)."""

    def test_requests_without_timeout_fix(self):
        from mcpaudit.rules.resource_exhaustion import ResourceExhaustionRule
        source = "r = requests.get(url)\n"
        rule = ResourceExhaustionRule()
        findings = _check(rule, source)
        assert len(findings) >= 1
        fix = rule.suggest_fix(source, findings[0])
        assert fix is not None
        assert "timeout" in fix.lower()

    def test_timeout_inserted_before_paren(self):
        from mcpaudit.rules.resource_exhaustion import ResourceExhaustionRule
        source = "r = requests.get(url)\n"
        rule = ResourceExhaustionRule()
        findings = _check(rule, source)
        assert len(findings) >= 1
        fix = rule.suggest_fix(source, findings[0])
        assert "timeout=30" in fix

    def test_empty_call_adds_timeout(self):
        from mcpaudit.rules.resource_exhaustion import ResourceExhaustionRule
        rule = ResourceExhaustionRule()
        finding = Finding(
            rule_id="RES-001", severity=Severity.MEDIUM,
            title="test", description="test",
            file="test.py", line=1,
        )
        source = "data = f.read()\n"
        fix = rule.suggest_fix(source, finding)
        assert fix is not None
        assert "timeout" in fix.lower()

    def test_invalid_line_returns_none(self):
        from mcpaudit.rules.resource_exhaustion import ResourceExhaustionRule
        rule = ResourceExhaustionRule()
        finding = Finding(
            rule_id="RES-001", severity=Severity.MEDIUM,
            title="test", description="test",
            file="test.py", line=999,
        )
        fix = rule.suggest_fix("x = 1\n", finding)
        assert fix is None


# ===========================================================================
# INFO-001: information disclosure
# ===========================================================================

class TestFixINFO001:
    """Tests for INFO-001 suggest_fix (information disclosure)."""

    def test_info_disclosure_fix(self):
        from mcpaudit.rules.info_disclosure import InformationDisclosureRule
        rule = InformationDisclosureRule()
        finding = Finding(
            rule_id="INFO-001", severity=Severity.LOW,
            title="test", description="test",
            file="test.py", line=1,
        )
        source = "except Exception as e:\n"
        fix = rule.suggest_fix(source, finding)
        assert fix is not None
        assert "generic" in fix.lower() or "TODO" in fix

    def test_invalid_line_returns_none(self):
        from mcpaudit.rules.info_disclosure import InformationDisclosureRule
        rule = InformationDisclosureRule()
        finding = Finding(
            rule_id="INFO-001", severity=Severity.LOW,
            title="test", description="test",
            file="test.py", line=999,
        )
        fix = rule.suggest_fix("x = 1\n", finding)
        assert fix is None


# ===========================================================================
# MCP-005: tool input validation
# ===========================================================================

class TestFixMCP005:
    """Tests for MCP-005 suggest_fix (input validation)."""

    def test_tool_handler_fix(self):
        from mcpaudit.rules.mcp_rules import MCPToolInputValidationRule
        source = (
            "async def call_tool(name, arguments):\n"
            "    result = process(arguments['query'])\n"
            "    return result\n"
        )
        rule = MCPToolInputValidationRule()
        findings = _check(rule, source)
        assert len(findings) >= 1
        fix = rule.suggest_fix(source, findings[0])
        assert fix is not None
        assert "isinstance" in fix or "TODO" in fix

    def test_invalid_line_returns_none(self):
        from mcpaudit.rules.mcp_rules import MCPToolInputValidationRule
        rule = MCPToolInputValidationRule()
        finding = Finding(
            rule_id="MCP-005", severity=Severity.HIGH,
            title="test", description="test",
            file="test.py", line=999,
        )
        fix = rule.suggest_fix("x = 1\n", finding)
        assert fix is None


# ===========================================================================
# MCP-006: response size limit
# ===========================================================================

class TestFixMCP006:
    """Tests for MCP-006 suggest_fix (response size)."""

    def test_response_size_fix(self):
        from mcpaudit.rules.mcp_rules import MCPToolResponseSizeRule
        source = (
            "async def call_tool(name, arguments):\n"
            "    data = f.read()\n"
            "    return data\n"
        )
        rule = MCPToolResponseSizeRule()
        findings = _check(rule, source)
        assert len(findings) >= 1
        fix = rule.suggest_fix(source, findings[0])
        assert fix is not None
        assert "MAX" in fix or "truncat" in fix.lower() or "TODO" in fix

    def test_invalid_line_returns_none(self):
        from mcpaudit.rules.mcp_rules import MCPToolResponseSizeRule
        rule = MCPToolResponseSizeRule()
        finding = Finding(
            rule_id="MCP-006", severity=Severity.MEDIUM,
            title="test", description="test",
            file="test.py", line=999,
        )
        fix = rule.suggest_fix("x = 1\n", finding)
        assert fix is None


# ===========================================================================
# CRYPTO-001: weak hash
# ===========================================================================

class TestFixCRYPTO001:
    """Tests for CRYPTO-001 rule and suggest_fix (weak hash)."""

    def test_detect_md5(self):
        from mcpaudit.rules.weak_crypto import WeakCryptoHashRule
        source = "h = hashlib.md5(data)\n"
        rule = WeakCryptoHashRule()
        findings = _check(rule, source)
        assert len(findings) == 1
        assert findings[0].rule_id == "CRYPTO-001"
        assert "MD5" in findings[0].title

    def test_detect_sha1(self):
        from mcpaudit.rules.weak_crypto import WeakCryptoHashRule
        source = "h = hashlib.sha1(data)\n"
        rule = WeakCryptoHashRule()
        findings = _check(rule, source)
        assert len(findings) == 1
        assert "SHA1" in findings[0].title

    def test_sha256_no_finding(self):
        from mcpaudit.rules.weak_crypto import WeakCryptoHashRule
        source = "h = hashlib.sha256(data)\n"
        rule = WeakCryptoHashRule()
        findings = _check(rule, source)
        assert len(findings) == 0

    def test_comment_skipped(self):
        from mcpaudit.rules.weak_crypto import WeakCryptoHashRule
        source = "# hashlib.md5 is bad\n"
        rule = WeakCryptoHashRule()
        findings = _check(rule, source)
        assert len(findings) == 0

    def test_fix_md5_to_sha256(self):
        from mcpaudit.rules.weak_crypto import WeakCryptoHashRule
        source = "h = hashlib.md5(data)\n"
        rule = WeakCryptoHashRule()
        findings = _check(rule, source)
        assert len(findings) >= 1
        fix = rule.suggest_fix(source, findings[0])
        assert fix is not None
        assert "sha256" in fix
        assert "md5" not in fix

    def test_fix_sha1_to_sha256(self):
        from mcpaudit.rules.weak_crypto import WeakCryptoHashRule
        source = "h = hashlib.sha1(data)\n"
        rule = WeakCryptoHashRule()
        findings = _check(rule, source)
        assert len(findings) >= 1
        fix = rule.suggest_fix(source, findings[0])
        assert fix is not None
        assert "sha256" in fix
        assert "sha1" not in fix

    def test_fix_invalid_line_returns_none(self):
        from mcpaudit.rules.weak_crypto import WeakCryptoHashRule
        rule = WeakCryptoHashRule()
        finding = Finding(
            rule_id="CRYPTO-001", severity=Severity.MEDIUM,
            title="test", description="test",
            file="test.py", line=999,
        )
        fix = rule.suggest_fix("x = 1\n", finding)
        assert fix is None

    def test_rule_properties(self):
        from mcpaudit.rules.weak_crypto import WeakCryptoHashRule
        rule = WeakCryptoHashRule()
        assert rule.rule_id == "CRYPTO-001"
        assert rule.name == "Weak Cryptographic Hash"
        assert rule.severity == Severity.MEDIUM
        assert rule.cwe == "CWE-328"

    def test_multiple_findings(self):
        from mcpaudit.rules.weak_crypto import WeakCryptoHashRule
        source = "a = hashlib.md5(x)\nb = hashlib.sha1(y)\n"
        rule = WeakCryptoHashRule()
        findings = _check(rule, source)
        assert len(findings) == 2

    def test_detailed_description(self):
        from mcpaudit.rules.weak_crypto import WeakCryptoHashRule
        rule = WeakCryptoHashRule()
        desc = rule.detailed_description
        assert "MD5" in desc
        assert "SHA1" in desc
        assert "sha256" in desc.lower()


# ===========================================================================
# FixStats
# ===========================================================================

class TestFixStats:
    """Tests for FixStats dataclass."""

    def test_empty_stats(self):
        stats = FixStats()
        assert stats.total_findings == 0
        assert stats.fixable_findings == 0
        assert stats.coverage_pct == 0.0
        assert stats.success_pct == 0.0

    def test_coverage_calculation(self):
        stats = FixStats(total_findings=100, fixable_findings=50)
        assert stats.coverage_pct == 50.0

    def test_success_calculation(self):
        stats = FixStats(fixable_findings=10, applied_fixes=8)
        assert stats.success_pct == 80.0

    def test_to_dict(self):
        stats = FixStats(
            total_findings=10,
            fixable_findings=5,
            applied_fixes=3,
            by_rule={"CMD-001": 2, "SQL-001": 1},
        )
        d = stats.to_dict()
        assert d["total_findings"] == 10
        assert d["fixable_findings"] == 5
        assert d["applied_fixes"] == 3
        assert d["coverage_pct"] == 50.0
        assert d["by_rule"]["CMD-001"] == 2


# ===========================================================================
# compute_fix_stats
# ===========================================================================

class TestComputeFixStats:
    """Tests for compute_fix_stats()."""

    def test_empty_results(self):
        stats = compute_fix_stats([], total_findings=0)
        assert stats.total_findings == 0
        assert stats.fixable_findings == 0

    def test_with_suggestions(self):
        finding = Finding(
            rule_id="CMD-001", severity=Severity.CRITICAL,
            title="test", description="test",
            file="test.py", line=1,
        )
        fr = FixResult(
            file="test.py",
            suggestions=[
                FixSuggestion(
                    finding=finding,
                    original_line="a",
                    fixed_line="b",
                    line_number=1,
                ),
            ],
        )
        stats = compute_fix_stats([fr], total_findings=5)
        assert stats.total_findings == 5
        assert stats.fixable_findings == 1
        assert stats.by_rule["CMD-001"] == 1

    def test_applied_fixes(self):
        finding = Finding(
            rule_id="SQL-001", severity=Severity.CRITICAL,
            title="test", description="test",
            file="test.py", line=1,
        )
        fr = FixResult(
            file="test.py",
            applied=True,
            suggestions=[
                FixSuggestion(
                    finding=finding,
                    original_line="a",
                    fixed_line="b",
                    line_number=1,
                ),
            ],
        )
        stats = compute_fix_stats([fr], total_findings=1)
        assert stats.applied_fixes == 1
        assert stats.files_modified == ["test.py"]
        assert stats.success_pct == 100.0

    def test_with_errors(self):
        fr = FixResult(file="broken.py", error="Cannot read file")
        stats = compute_fix_stats([fr], total_findings=0)
        assert stats.errors == 1

    def test_multiple_files(self):
        finding1 = Finding(
            rule_id="CMD-001", severity=Severity.CRITICAL,
            title="test", description="test",
            file="a.py", line=1,
        )
        finding2 = Finding(
            rule_id="SQL-001", severity=Severity.CRITICAL,
            title="test", description="test",
            file="b.py", line=1,
        )
        fr1 = FixResult(
            file="a.py",
            applied=True,
            suggestions=[FixSuggestion(
                finding=finding1, original_line="a", fixed_line="b", line_number=1,
            )],
        )
        fr2 = FixResult(
            file="b.py",
            applied=True,
            suggestions=[FixSuggestion(
                finding=finding2, original_line="c", fixed_line="d", line_number=1,
            )],
        )
        stats = compute_fix_stats([fr1, fr2], total_findings=10)
        assert stats.fixable_findings == 2
        assert stats.applied_fixes == 2
        assert len(stats.files_modified) == 2
        assert stats.by_rule == {"CMD-001": 1, "SQL-001": 1}


# ===========================================================================
# format_fix_stats
# ===========================================================================

class TestFormatFixStats:
    """Tests for format_fix_stats()."""

    def test_empty_stats(self):
        stats = FixStats()
        output = format_fix_stats(stats)
        assert "Auto-Fix Statistics" in output
        assert "0" in output

    def test_with_data(self):
        stats = FixStats(
            total_findings=20,
            fixable_findings=10,
            applied_fixes=8,
            by_rule={"CMD-001": 5, "SQL-001": 3},
        )
        output = format_fix_stats(stats)
        assert "Total findings:    20" in output
        assert "Fixable findings:  10" in output
        assert "CMD-001" in output
        assert "SQL-001" in output

    def test_shows_errors(self):
        stats = FixStats(errors=2)
        output = format_fix_stats(stats)
        assert "Errors:            2" in output


# ===========================================================================
# End-to-end fix integration
# ===========================================================================

class TestEndToEndFix:
    """End-to-end tests for the enhanced fix engine."""

    def test_fix_crypto_md5_file(self):
        source = "import hashlib\nh = hashlib.md5(data)\n"
        f = _write_temp(source)
        from mcpaudit.rules.weak_crypto import WeakCryptoHashRule
        rule = WeakCryptoHashRule()
        findings = _check(rule, source)
        for fi in findings:
            fi.file = str(f)
        rm = build_rule_map([rule])
        result = suggest_fixes(str(f), findings, rm)
        assert result.fixable_count >= 1
        ok = apply_fixes(result)
        assert ok is True
        content = Path(f).read_text()
        assert "sha256" in content

    def test_fix_yaml_load_file(self):
        source = "import yaml\ndata = yaml.load(f)\n"
        f = _write_temp(source)
        from mcpaudit.rules.unsafe_deser import UnsafeDeserializationRule
        rule = UnsafeDeserializationRule()
        findings = _check(rule, source)
        for fi in findings:
            fi.file = str(f)
        rm = build_rule_map([rule])
        result = suggest_fixes(str(f), findings, rm)
        assert result.fixable_count >= 1
        ok = apply_fixes(result)
        assert ok is True
        content = Path(f).read_text()
        assert "safe_load" in content

    def test_fix_shell_true_file(self):
        source = "import subprocess\nsubprocess.run('ls', shell=True)\n"
        f = _write_temp(source)
        from mcpaudit.rules.command_injection import CommandInjectionShellRule
        rule = CommandInjectionShellRule()
        findings = _check(rule, source)
        for fi in findings:
            fi.file = str(f)
        rm = build_rule_map([rule])
        result = suggest_fixes(str(f), findings, rm)
        assert result.fixable_count >= 1
        ok = apply_fixes(result)
        assert ok is True
        content = Path(f).read_text()
        assert "shell=False" in content

    def test_stats_from_multi_rule_scan(self):
        """Test statistics from scanning with multiple rules."""
        source = (
            "import hashlib\n"
            "h = hashlib.md5(data)\n"
            "import subprocess\n"
            "subprocess.run('ls', shell=True)\n"
        )
        f = _write_temp(source)
        rules = get_all_rules()
        rm = build_rule_map(rules)

        from mcpaudit.scanner import MCPSecurityScanner
        scanner = MCPSecurityScanner(rules=rules)
        scan_result = scanner.scan_file(Path(f))

        for fi in scan_result.findings:
            fi.file = str(f)

        result = suggest_fixes(str(f), scan_result.findings, rm)
        stats = compute_fix_stats([result], total_findings=len(scan_result.findings))
        assert stats.fixable_findings >= 2
        assert stats.coverage_pct > 0


# ===========================================================================
# CLI integration
# ===========================================================================

class TestFixCLIEnhanced:
    """Tests for enhanced CLI fix flags."""

    def test_fix_stats_flag_registered(self):
        from mcpaudit.__main__ import build_parser
        parser = build_parser()
        args = parser.parse_args(["test.py", "--fix-stats"])
        assert args.fix_stats is True

    def test_fix_stats_output(self, tmp_path: Path, capsys):
        f = tmp_path / "test.py"
        f.write_text("import hashlib\nh = hashlib.md5(data)\n")
        from mcpaudit.__main__ import main
        exit_code = main([str(f), "--fix-stats"])
        captured = capsys.readouterr()
        assert "Auto-Fix Statistics" in captured.out

    def test_fix_dry_run_crypto(self, tmp_path: Path, capsys):
        f = tmp_path / "test.py"
        f.write_text("import hashlib\nh = hashlib.md5(data)\n")
        from mcpaudit.__main__ import main
        exit_code = main([str(f), "--fix-dry-run"])
        captured = capsys.readouterr()
        assert "Auto-Fix Preview" in captured.out

    def test_fix_applies_crypto(self, tmp_path: Path, capsys):
        f = tmp_path / "test.py"
        f.write_text("import hashlib\nh = hashlib.md5(data)\n")
        from mcpaudit.__main__ import main
        exit_code = main([str(f), "--fix"])
        captured = capsys.readouterr()
        assert "Auto-Fix Applied" in captured.out
        content = f.read_text()
        assert "sha256" in content
